from inspect import isclass
from collections import OrderedDict
import copy
from .exceptions import OrmParameterError, OrmModelDeclarationError, OrmManagementError


class Mapper(object):
    def __init__(self):
        self.models = dict()
        self.templates = dict()
        self.models_declared = list()
        self.templates_models = dict()
        self.models_templates = dict()
        self.userident_model = None
        self.many2many = OrderedDict()
        self.m2m_map = {}
        self.m2m_c = {}
        self._on_model_declare = dict()

    def validate_all(self):
        for model_name in self.models:
            model = self.models[model_name]
            model_attrs = model.__meta__.get_attributes()
            for attr in model_attrs:
                if not hasattr(attr, 'validate'):
                    continue
                attr.validate()

    def map_template(self, template):
        self.templates[template.__name__] = template

    def has_template(self, name):
        return name in self.templates

    def get_template_by_name(self, name):
        return None if name not in self.templates else self.templates[name]

    def unmap_template(self, name):
        if name not in self.templates:
            return
        if name in self.templates_models:
            raise OrmManagementError("cannot unmap template '%s' because it is already used in Models!" % name)
        del(self.templates[name])

    def is_model_templated_with(self, model, template):
        model_ = model if isinstance(model, str) else model.__name__
        template_ = template if isinstance(template, str) else template.__name__
        if not self.has_template(template_):
            return False
        return template_ in self.models_templates[model_]

    def map_model(self, model):
        self.models[model.__name__] = model

    def has_model(self, name):
        return name in self.models

    def unmap_model(self, name):
        if name not in self.models:
            return
        del(self.models[name])
        if name in self.models_declared:
            del(self.models_declared[self.models_declared.index(name)])
        if name in self.models_templates:
            for template in self.models_templates[name]:
                if template not in self.templates_models or name not in self.templates_models[template]:
                    continue
                del(self.templates_models[template][self.templates_models[template].index(name)])

    def get_model_by_name(self, name):
        return None if name not in self.models else self.models[name]

    def get_models_for_migrate(self, tag=None):
        for_migrate = dict()
        for k in self.models:
            if self.models[k].__meta__.tag is not None and self.models[k].__meta__.tag != tag:
                continue
            for_migrate[k] = self.models[k]
        return for_migrate

    def model_declare(self, model_name):
        if model_name not in self.models_declared:
            self.models_declared.append(model_name)
        self.call_on_model_declare(model_name)

    def register_on_model_declare(self, model_name, callback, *args, **kwargs):
        if model_name not in self._on_model_declare:
            self._on_model_declare[model_name] = list()
        self._on_model_declare[model_name].append((callback, args, kwargs))
        if model_name in self.models_declared:
            self.call_on_model_declare(model_name)

    def call_on_model_declare(self, model_name):
        if model_name not in self._on_model_declare or not self._on_model_declare[model_name]:
            return
        for trigger in self._on_model_declare[model_name]:
            callback = trigger[0]
            args = trigger[1]
            kwargs = trigger[2]
            callback(*args, **kwargs)
        self._on_model_declare[model_name] = list()

    def is_model_declared(self, model_name):
        return model_name in self.models

    def register_templated_model(self, template, model):
        if not isinstance(template, str):
            template = template.__name__
        if not isinstance(model, str):
            model = model.__name__
        if template not in self.templates_models:
            self.templates_models[template] = list()
        if model not in self.templates_models[template]:
            self.templates_models[template].append(model)
        if model not in self.models_templates:
            self.models_templates[model] = list()
        if template not in self.models_templates[model]:
            self.models_templates[model].append(template)

    def is_model_templated(self, model):
        if not isinstance(model, str):
            model = model.__name__
        return model in self.models_templates

    def is_template_used(self, template):
        if not isinstance(template, str):
            template = template.__name__
        return template in self.templates_models

    def get_template_models(self, template):
        if not isinstance(template, str):
            template = template.__name__
        return None if template not in self.templates_models else self.templates_models[template]

    def get_model_templates(self, model):
        if not isinstance(model, str):
            model = model.__name__
        return None if not model or model not in self.models_templates else self.models_templates[model]

    def set_userident_model(self, model):
        self.userident_model = model

    def get_userident_model(self):
        return self.userident_model

    def is_userident_model_set(self):
        return bool(self.userident_model is not None)

    def _many2many_create(self, key):
        from .models.base import OrmModel
        from .attributes.base import ColumnAttributeAbstract
        from .attributes.relationships import ForeignKey

        m2m_c = self.m2m_c[key][0]
        fk_ondelete = m2m_c.on_delete
        model_name = "%s_2_%s" % (key[0], key[1])
        table_name = model_name.lower()
        model_a = self.get_model_by_name(key[0])
        model_b = self.get_model_by_name(key[1])
        pk_a = model_a.__meta__.get_primary_key()
        pk_b = model_b.__meta__.get_primary_key()
        if not pk_a or not pk_b:
            raise OrmModelDeclarationError(
                "'ManyToMany' requires that both first and second models have primary key declared!"
            )
        m2m_attributes = {
            '__prefix__': None,
            '__tablename__': table_name
        }
        decl_order = 0
        fk_a_l = list()
        fk_a_r = list()
        fk_b_l = list()
        fk_b_r = list()
        for c in pk_a:
            k_ = "%s_%s" % (str(key[0]).lower(), c.model_k)
            attribute_class = c.base_type or c.__class__
            c_ = attribute_class()
            c_.__dict__ = copy.deepcopy(c.__dict__)
            c_.model_k = k_
            c_.tbl_k = k_
            c_.primary_key = True
            c_.auto_increment = False
            c_.nullable = False
            c_.default = None
            c_.parent_model_k = c.model_k
            c_.parent_tbl_k = c.tbl_k
            c_.parent_model = c.model
            c_.declaration_order = decl_order
            m2m_attributes[k_] = c_
            fk_a_l.append(c_)
            fk_a_r.append(c)
            decl_order += 10
        for c in pk_b:
            k_ = "%s_%s" % (str(key[1]).lower(), c.model_k)
            attribute_class = c.base_type or c.__class__
            c_ = attribute_class()
            c_.__dict__ = copy.deepcopy(c.__dict__)
            c_.model_k = k_
            c_.tbl_k = k_
            c_.primary_key = True
            c_.auto_increment = False
            c_.nullable = False
            c_.default = None
            c_.parent_model_k = c.model_k
            c_.parent_tbl_k = c.tbl_k
            c_.parent_model = c.model
            c_.declaration_order = decl_order
            m2m_attributes[k_] = c_
            fk_b_l.append(c_)
            fk_b_r.append(c)
            decl_order += 10
        fk_a = ForeignKey(fk_a_l, fk_a_r, on_delete=fk_ondelete)
        fk_b = ForeignKey(fk_b_l, fk_b_r, on_delete=fk_ondelete)
        m2m_attributes['ibfk_a_'] = fk_a
        m2m_attributes['ibfk_b_'] = fk_b
        m2m_model = type(model_name, (OrmModel,), m2m_attributes)
        m2m_model.__m2m_columns__ = {
            key[0]: [],
            key[1]: []
        }
        attrs = dir(m2m_model)
        for k in attrs:
            if k.startswith('__') and k.endswith('__'):
                continue
            c = getattr(m2m_model, k)
            if not isinstance(c, ColumnAttributeAbstract):
                continue
            if c.parent_model == model_a:
                m2m_model.__m2m_columns__[key[0]].append(c)
            else:
                m2m_model.__m2m_columns__[key[1]].append(c)
        self.many2many[key] = m2m_model
        for c in self.m2m_c[key]:
            c.validate(m2m_model)

    def _many2many_on_model_declared(self, model):
        model_name = model.__name__ if not isinstance(model, str) else model
        if model_name not in self.m2m_map:
            return
        key = self.m2m_map[model_name]
        if not self.is_model_declared(key[0]) or not self.is_model_declared(key[1]):
            return
        self._many2many_create(key)

    def register_many2many(self, c, model_a, model_b):
        m2m_key = (model_a, model_b)
        if m2m_key not in self.m2m_c:
            self.m2m_c[m2m_key] = list()
        if c not in self.m2m_c[m2m_key]:
            self.m2m_c[m2m_key].append(c)
        if m2m_key in self.many2many:
            return m2m_key
        self.many2many[m2m_key] = None
        self.m2m_map[model_a] = m2m_key
        self.m2m_map[model_b] = m2m_key
        self.register_on_model_declare(model_a, self._many2many_on_model_declared, model_a)
        self.register_on_model_declare(model_b, self._many2many_on_model_declared, model_b)
        return m2m_key

    def get_many2many_junction_model(self, key):
        if key not in self.many2many:
            return None
        return self.many2many[key]

    def get_many2many_model_by_c(self, c):
        from .attributes.relationships import ManyToMany
        if not isinstance(c, ManyToMany):
            raise OrmParameterError(
                "'mapper'.'get_many2many_model_by_c' requires ManyToMany relationship attribute to be given!"
            )
        key = (c.model_a, c.model_b)
        return self.get_many2many_junction_model(key)

    @staticmethod
    def define_attribute_model(parent_model, model_k, attributes):
        from .models.base import OrmModel
        from .attributes.derived import BigAutoIncrement
        from .attributes.relationships import ForeignKey, FK_ONDELETE_CASCADE
        model_name = "%s_%s" % (parent_model.__name__, model_k)
        table_name = model_name.lower()
        model_attrs = {
            '__prefix__': None,
            '__tablename__': table_name,
            'id': BigAutoIncrement()
        }
        model_attrs['id'].declaration_order = 0
        decl_order = 10
        parent_pk = parent_model.__meta__.get_primary_key()
        fk_l = list()
        fk_r = list()
        for c in parent_pk:
            k_ = "parent_%s" % c.model_k
            attribute_class = c.base_type or c.__class__
            c_ = attribute_class()
            c_.__dict__ = copy.deepcopy(c.__dict__)
            c_.model_k = k_
            c_.tbl_k = k_
            c_.primary_key = False
            c_.auto_increment = False
            c_.nullable = False
            c_.default = None
            c_.parent_model_k = c.model_k
            c_.parent_tbl_k = c.tbl_k
            c_.parent_model = c.model
            c_.declaration_order = decl_order
            model_attrs[k_] = c_
            fk_l.append(c_)
            fk_r.append(c)
            decl_order += 10
        for k in attributes:
            c = attributes[k]
            model_attrs[k] = c() if isclass(c) else c
        fk = ForeignKey(fk_l, fk_r, on_delete=FK_ONDELETE_CASCADE)
        model_attrs['ibfk_parent'] = fk
        return type(model_name, (OrmModel,), model_attrs)


mapper = Mapper()



import datetime
from pc23 import metaclass, xrange
from .meta import OrmModelMeta
from ..attributes.base import OrmComparatorMixin, DataAttributeAbstract, AttributeAbstract, ColumnAttributeAbstract
from ..attributes.virtual import ArrayAbstract
from ..values.base import wrap_value, AttributeValueVirtual, AttributeValueNull, AttributeValueAbstract
from ..attributes.relationships import ForeignAttr, RelatedToOneAbstract, RelatedToManyAbstract, ManyToMany, OneToMany
from ..funcs import DbFunctionAbstract, RequestTableBelongedColumn
from ..exceptions import OrmModelDeclarationError, OrmInstanceMustBeWritten, OrmParameterError
from ..results import OrmRelatedObjectNotAssigned, OrmResultOneToMany, OrmModelUpdateAction, OrmResultManyToMany
from ..manage import manager
from ..logs import OrmJournal


class OrmModelAbstract(object):
    def __new__(cls, *args, **kwargs):
        instance = super(OrmModelAbstract, cls).__new__(cls, *args)
        instance.__session__ = None
        instance.__existing__ = False
        instance.__parent__ = None
        instance.__m2m__ = set()
        instance.__changes__ = dict()
        instance.__triggers__ = True
        instance.__attributes__ = dict()
        instance.__loadedkeys__ = list()
        instance.log = OrmJournal(cls, instance)
        return instance

    @property
    def __modified__(self):
        """ Returns True if Model's instance been modified (and need to be written to the database to
        save changed), and False if nothing been changed or all changes been cancelled by setting their
        values to the original ones. """
        return bool(self.__changes__)

    def __setattrvalue__(self, key, value):
        """ Sets the value for the Model instance's attribute. This method uses logics of the corresponding
        attibute to store given value, processing the value via attribute's methods and resulting end-point
        one may differ from the origial given. """
        if key not in self.__attributes__ and not hasattr(self.__class__, key):
            raise AttributeError(str(key))
        c, model = self.__attributes__[key] \
            if key in self.__attributes__ \
            else getattr(self.__class__, key), self.__class__
        if not isinstance(c, (DataAttributeAbstract, DbFunctionAbstract)):
            super(OrmModelAbstract, self).__setattr__(key, value)
            return
        value_new = wrap_value(model, self, c, key, value)
        value_old = super(OrmModelAbstract, self).__getattribute__(key)
        if not isinstance(value_old, OrmComparatorMixin) and value_old == value_new:
            return
        super(OrmModelAbstract, self).__setattr__(key, value_new)

    def __setattrabs__(self, key, value):
        """ Sets the instance's attribute to the given value, bypassing any Model attribute's logics.
        @key:               The name of the attribute to set to;
        @value:             The value to set for the instance's attribute;
        """
        super(OrmModelAbstract, self).__setattr__(key, value)

    def __getforwrite__(self, key, session):
        """ Returns a value which can be passed to the database connector library as value for
        write (INSERT or UPDATE). """
        value = self.__getattribute__(key)
        if isinstance(value, AttributeValueNull):
            value = None
        elif isinstance(value, AttributeValueVirtual):
            value = value.dbvalue if hasattr(value, 'dbvalue') else value.value
        c = getattr(self.__class__, key, None)
        if isinstance(c, ColumnAttributeAbstract):
            return c.represent_for_write(value, session)
        return value

    def __initialize_from__(self, **kwargs):
        """ Initializes this instance's attributes from initial data defined for the Model. Using
        when migrating database schema in case when table creates first time. """
        triggers_enabled = self.__triggers__
        self.__triggers__ = False
        for k in kwargs:
            if k.startswith('__') and k.endswith('__'):
                continue
            c = getattr(self.__class__, k, None)
            if not c:
                setattr(self, k, kwargs[k])
                continue
            value = kwargs[k]
            if isinstance(value, (list, tuple)) and not getattr(c, 'enumerable_value', False):
                childs = getattr(self, k)
                if not isinstance(childs, (OrmResultOneToMany, OrmResultManyToMany)):
                    raise OrmModelDeclarationError(
                        "__initial__ value declared as an array must point to the ..ToMany relationship!"
                    )
                if isinstance(childs, OrmResultManyToMany):
                    continue
                for ci in value:
                    if not isinstance(ci, dict):
                        raise OrmParameterError(
                            "__initial__ values must be a type of list of dicts for instance itself and child ones!"
                        )
                    childs.add(**ci)
                continue
            setattr(self, k, value)
        self.__triggers__ = triggers_enabled
        self.__afterload__()

    def __load__(self, **kwargs):
        """ Loads initial values, loaded from the database by SELECT, to the Model's instance. """
        triggers_enabled = self.__triggers__
        self.__triggers__ = False
        textual = kwargs.pop('__textual__', False)
        for k in kwargs:
            d = kwargs[k]
            value = d[0] if isinstance(d, (list, tuple)) else d
            c = d[1] \
                if isinstance(d, (list, tuple)) and len(d) > 1 \
                else (self.__attributes__[k] if k in self.__attributes__ else getattr(self.__class__, k))
            model = d[2] \
                if isinstance(d, (list, tuple)) and len(d) > 2 \
                else (self.__class__ if CompositeOrmModel not in self.__class__.mro() else None)
            value = wrap_value(model, self, c, k, value, textual)
            self.__setattr__(k, value, on_instance_loaded=True)
            if k not in self.__attributes__ and model is not None:
                self.__attributes__[k] = c, model
            if k not in self.__loadedkeys__:
                self.__loadedkeys__.append(k)
        self.__triggers__ = triggers_enabled
        self.__afterload__()

    def __afterload__(self):
        """ Called after instance been created and all values been loaded. """
        triggers_enabled = self.__triggers__
        self.__triggers__ = False
        attrs = self.__attributes__
        for k in attrs:
            c = attrs[k][0]
            c.on_instance_load(self)
        self.__triggers__ = triggers_enabled

    def __check_modified__(self, key, new_value, *args):
        """ Verifies was the attribute changed or not by testing given values, comparing 'new_value' with
        the original one (original means value which been loaded from the database or been set at a time
        of instance creating). The original value may be proposed as last optional argument (the first
        argument of *args).
        This method returns nothing, but handles instance's __changed__ dict, adding a record about changed
        attribute to it if given (by name via 'key' argument) attribute been changed, and deleting a
        corresponding record if the original value of the attribute and new one - are same.
        @key:                   The name of the Model's attribute to verify to;
        @new_value:             The new value which is about to be set to the attribute at this time;
        @*args:                 (Optional) if given - the first argument of 'args' will be used as old
                                value instead of reading the actual value from the instance; otherwise
                                old value (to compare with) will be readen from the instance.
        """
        original_value = self.__changes__[key][0] \
            if key in self.__changes__ \
            else (args[0] if args else self.__getattribute__(key))
        new_value = new_value if not isinstance(new_value, AttributeValueNull) else None
        original_value = original_value if not isinstance(original_value, AttributeValueNull) else None
        if isinstance(new_value, AttributeValueAbstract):
            new_value = new_value.value
        if isinstance(original_value, AttributeValueAbstract):
            original_value = original_value.value
        is_changed = new_value != original_value
        if is_changed:
            self.__changes__[key] = (original_value, new_value)
        elif key in self.__changes__:
            del (self.__changes__[key])
        meta = getattr(self.__class__, '__meta__', None)
        if meta is not None and is_changed and getattr(meta, 'on_change', None) is not None:
            meta.on_change(self)

    def __set_modified__(self, key, old_value, new_value):
        """ Force set the record that attribute with given name been changed, not
        testing this fact.
        @key:                   The name of the attribute;
        @old_value:             Old value of the attribute;
        @new_value:             New value of the attribute;
        """
        self.__changes__[key] = (old_value, new_value)

    def __reset_modified__(self, key):
        """ Force remove the record (if any set before) that attribute with given name been changed,
        not testing that fact. Even if attribute's value been changed, the instance will not know
        about it any more. """
        if key not in self.__changes__:
            return
        del(self.__changes__[key])

    def __getattrabs__(self, key):
        """ Returns raw value of the instance's attribute, bypassing any attribute logics. """
        return super(OrmModelAbstract, self).__getattribute__(key)

    def __setattr__(self, key, value, on_instance_loaded=False):
        if key.startswith('__') and key.endswith('__'):
            super(OrmModelAbstract, self).__setattr__(key, value)
            return value
        if key not in dir(self.__class__) and key not in self.__attributes__:
            super(OrmModelAbstract, self).__setattr__(key, value)
            return value
        c = self.__attributes__[key][0] if key in self.__attributes__ else getattr(self.__class__, key)
        if isinstance(c, RequestTableBelongedColumn):
            c = getattr(self.__class__, key, None) or c
        if isinstance(c, AttributeAbstract) and c.setter is not None:
            original_value_ = self.__getattribute__(key)
            value_ = c.setter(self, key, value, bool(on_instance_loaded))
            if self.__triggers__:
                self.__check_modified__(key, value_, original_value_)
                c.call_on_change(self, key, value_)
            return value_
        iv = self.__getattrabs__(key)
        if isinstance(iv, AttributeValueVirtual):
            value_ = iv.set_value(self, key, value)
            if self.__triggers__:
                if iv.check_modified is not False:
                    if isinstance(value_, (list, tuple)):
                        self.__check_modified__(key, *value_)
                    else:
                        self.__check_modified__(key, value_)
                if isinstance(c, AttributeAbstract):
                    c.call_on_change(self, key, value_)
            return value
        if isinstance(c, DataAttributeAbstract):
            if self.__triggers__:
                value = c.verify(value)
                self.__check_modified__(key, value)
            self.__setattrvalue__(key, value)
            if self.__triggers__:
                c.call_on_change(self, key, value)
            return value
        super(OrmModelAbstract, self).__setattr__(key, value)
        if self.__triggers__ and isinstance(c, AttributeAbstract):
            c.call_on_change(self, key, value)
        return value

    def __getattribute__(self, key):
        if isinstance(key, str) and key.startswith('__') and key.endswith('__'):
            return super(OrmModelAbstract, self).__getattribute__(key)
        if key not in dir(self.__class__) and key not in self.__attributes__:
            return super(OrmModelAbstract, self).__getattribute__(key)
        c = self.__attributes__[key][0] if key in self.__attributes__ else getattr(self.__class__, key)
        if isinstance(c, AttributeAbstract):
            c.validate()
        if isinstance(c, AttributeAbstract) and c.getter is not None:
            return c.getter(self, key)
        value = super(OrmModelAbstract, self).__getattribute__(key)
        if isinstance(value, AttributeValueVirtual):
            value = value.get_value()
        if isinstance(value, OrmRelatedObjectNotAssigned):
            value.__parent_k__ = key
        return value


@metaclass(OrmModelMeta)
class OrmModel(OrmModelAbstract):
    def __new__(cls, *args, **kwargs):
        """ Overrides default __new__ method of the Model's class, executing some additional
        actions in addition to default ones. """
        instance = super(OrmModel, cls).__new__(cls, *args)
        attr_keys = dir(cls)
        for k in attr_keys:
            if k.startswith('__') and k.endswith('__'):
                continue
            c = getattr(cls, k)
            if isinstance(c, AttributeAbstract):
                if k not in kwargs \
                        and isinstance(c, DataAttributeAbstract) \
                        and not isinstance(c, ArrayAbstract) \
                        and 'default' in dir(c):
                    _default_value = getattr(c, 'default')
                    if _default_value is not None:
                        kwargs[k] = _default_value
                c.initialize(instance, k)
                if c.on_new is not None:
                    c.on_new(instance, k)
                continue
        if kwargs:
            for k in kwargs:
                if k.startswith('__') and k.endswith('__'):
                    continue
                instance.__load__(**kwargs)
        if cls.__meta__.on_instance_create is not None:
            cls.__meta__.on_instance_create(instance)
        return instance

    @property
    def __caption__(self):
        """ Returns caption of this Model's instance, if any caption attribute is defined. Returns
        an empty string if no caption attributes defined. """
        k_caption = self.__class__.__meta__.caption_key
        if k_caption is not None:
            return str(getattr(self, k_caption))
        return ""

    def __repr__(self):
        k_caption = self.__class__.__meta__.caption_key
        if k_caption is not None:
            return "%s(%r)" % (self.__class__.__name__, getattr(self, k_caption))
        return super(OrmModel, self).__repr__()

    def __init__(self, **kwargs):
        for k in kwargs:
            setattr(self, k, kwargs[k])

    def __afterload__(self):
        if self.__class__.__meta__.on_instance_loaded is not None:
            self.__class__.__meta__.on_instance_loaded(self)
        triggers_enabled = self.__triggers__
        self.__triggers__ = False
        attrs = self.__class__.__meta__.get_attributes()
        for c in attrs:
            c.on_instance_load(self)
        self.__triggers__ = triggers_enabled

    @property
    def dict(self):
        return self.__dictpair__

    @property
    def __dictpair__(self):
        k_caption = self.__class__.__meta__.caption_key
        if not k_caption:
            raise OrmModelDeclarationError(
                "Cannot convert the OrmModel's object instance to dict-pair (key and value) because OrmModel has"
                " no caption key defined. Programmer may use 'CaptionString' attribute type or 'is_caption_key'"
                " parameter to identify - which attribute to use as caption!"
            )
        k_value = self.__class__.__meta__.value_key
        if not k_value and len(self.__class__.__meta__.primary_key) > 1:
            raise OrmModelDeclarationError(
                "Cannot convert the OrmModel's object instance to dict-pair (key and value) because OrmModel has"
                " no value key defined and the primary key for this OrmModel is complex (only one column can"
                " be used as value key). By default, simple primary key is used as value key, or programmer"
                " must define a single column as value key by him(her)self!"
            )
        elif not k_value:
            k_value = self.__class__.__meta__.primary_key[0]
        return self.__getattribute__(k_value), self.__getattribute__(k_caption)

    @classmethod
    def as_(cls, name):
        """ Returns this Model formatted to be named Model as 'name'. """
        return OrmModelAliased(cls, name)

    @classmethod
    def get(cls, *args, **kwargs):
        from ..request import OrmSelectRequest
        return OrmSelectRequest(cls, *args, **kwargs).go()

    @classmethod
    def set(cls, session=None, **kwargs):
        return OrmModelUpdateAction(model=cls, session=session, values=kwargs)

    def describe(self, virtual=True, foreign=False, recursive=False):
        keys = self.__class__.__meta__.get_attributes_keys()
        p = list()
        p.append(self.__repr__())
        for k in keys:
            if k.startswith('__') and k.endswith('__'):
                continue
            c = self.__attributes__.get(k, None) or getattr(self.__class__, k, None)
            if isinstance(c, (list, tuple)):
                c = c[0]
            if not c:
                continue
            if not isinstance(c, AttributeAbstract):
                continue
            if isinstance(c, ForeignAttr) and not foreign:
                p.append("    %15s = (Foreign(%s.%s))" % (k, c.related_k, c.referenced_k))
                continue
            if isinstance(c, DataAttributeAbstract):
                if not virtual and isinstance(c, AttributeValueVirtual):
                    p.append("    %15s = (Virtual)" % k)
                    continue
                p.append("    %15s = \"%r\" (%s)" % (k, getattr(self, k), c.__class__.__name__))
                continue
            if isinstance(c, RelatedToOneAbstract):
                p.append("    %15s = (...ToOne relationship)" % k)
                continue
            if isinstance(c, RelatedToManyAbstract):
                p.append("    %15s = (...ToMany relationship)" % k)
                continue
        return "\n".join(p)

    def write(self, force=False, all_attrs=False, session=None, attrs=None, initial=False):
        from ..request import OrmInsertRequest, OrmUpdateRequest
        if initial:
            all_attrs = True
        rq_session = session or self.__session__ or manager.get_session(instance=self)
        setattr(self, '__session__', rq_session)
        if self.__class__.__meta__.on_before_write is not None \
                and self.__class__.__meta__.on_before_write(self) is False:
            return
        if self.__parent__ is not None:
            parent_instance, relation_c = self.__parent__
            for x in xrange(len(relation_c.columns)):
                instance_k = relation_c.columns[x]
                parent_k = relation_c.ref_columns[x]
                if relation_c.ref_model == self.__class__:
                    instance_k, parent_k = parent_k, instance_k
                parent_v = getattr(parent_instance, parent_k)
                setattr(self, instance_k, parent_v)
            self.log.replace_uuid(parent_instance.log.current_uuid)
        k_related = self.__class__.__meta__.get_related_loaded_keys(self)
        for k in k_related:
            c = getattr(self.__class__, k, None)
            if c is None or not isinstance(c, RelatedToManyAbstract):
                continue
            v = self.__getattribute__(k)
            if isinstance(v, OrmRelatedObjectNotAssigned):
                continue
            if not isinstance(v, OrmModel):
                continue
            if v.__existing__:
                continue
            v.write(session=rq_session, force=force)
        if force or self.__modified__ or not self.__existing__:
            self.__class__.__meta__.attrs_on_before_write(self)
            rq_cls = OrmUpdateRequest if self.__existing__ else OrmInsertRequest
            if self.__existing__ and self.__class__.__meta__.on_update is not None \
                    and self.__class__.__meta__.on_update(self) is False:
                return
            elif not self.__existing__ and self.__class__.__meta__.on_insert is not None \
                    and self.__class__.__meta__.on_insert(self) is False:
                return
            session_user = rq_session.inqure_session_user()
            if session_user and self.__class__.__meta__.author_ident and not self.__existing__:
                for c in self.__class__.__meta__.author_ident:
                    if getattr(self, c.ident_k, None) is not None:
                        continue
                    setattr(self, c.ident_k, session_user)
            if session_user and self.__class__.__meta__.performer_ident:
                for c in self.__class__.__meta__.performer_ident:
                    setattr(self, c.ident_k, session_user)
            if self.__class__.__meta__.timestamp_creation and not self.__existing__:
                for c in self.__class__.__meta__.timestamp_creation:
                    setattr(self, c.model_k, datetime.datetime.now())
            if self.__class__.__meta__.timestamp_modification:
                for c in self.__class__.__meta__.timestamp_modification:
                    setattr(self, c.model_k, datetime.datetime.now())
            rq_values = dict()
            write_all_attrs = not attrs and (all_attrs or force or not self.__existing__)
            rq_k_ = attrs \
                or (self.__class__.__meta__.get_keys_for_write() if write_all_attrs else self.__changes__.keys())
            rq_k = list()
            for k in rq_k_:
                if not hasattr(self.__class__, k):
                    continue
                c = getattr(self.__class__, k)
                c_attributes_ = getattr(c, 'attributes', None)
                if c_attributes_ is not None and isinstance(c_attributes_, (list, tuple)):
                    rq_k.extend(c_attributes_)
                if not isinstance(c, ColumnAttributeAbstract):
                    continue
                rq_k.append(k)
            for k in rq_k:
                c = getattr(self.__class__, k)
                v = self.__getforwrite__(k, rq_session)
                rq_values[c] = v
            if rq_values:
                rq = rq_cls(self.__class__, values=rq_values, with_deleted=True, session=rq_session)
                if self.__existing__:
                    w = list()
                    pk = self.__class__.__meta__.primary_key
                    for k in pk:
                        v = self.__getforwrite__(k, rq_session)
                        c = getattr(self.__class__, k)
                        w.append(c == v)
                    rq.where(*w)
                rq.go()
            else:
                rq = False
            # rq_session = rq.session
            k_autoincrement = self.__class__.__meta__.auto_increment
            if not self.__existing__ \
                    and k_autoincrement is not None \
                    and not self.__getattribute__(k_autoincrement)\
                    and rq is not False:
                self.__setattrabs__(k_autoincrement, rq.inserted_id)
            self.__setattrabs__('__session__', rq_session)
            self.log.write_submition()
            self.__setattrabs__('__changes__', dict())
            self.__setattrabs__('__existing__', True)
            vkeys = self.__class__.__meta__.get_virtualkeys_for_write(self)
            for k in vkeys:
                c = getattr(self.__class__, k)
                c.write(self, k)
            self.__class__.__meta__.attrs_on_after_write(self)
            if self.__class__.__meta__.on_after_write is not None:
                self.__class__.__meta__.on_after_write(self)
        for k in k_related:
            c = getattr(self.__class__, k, None)
            if c is not None and isinstance(c, ManyToMany) and initial:
                continue
            v = self.__getattribute__(k)
            if not hasattr(v, 'write'):
                continue
            if v is self:
                continue
            if isinstance(v, OrmRelatedObjectNotAssigned):
                continue
            v.write(session=rq_session, force=force)
        self.log.new_uuid()

    def delete(self, session=None, hard_delete=False):
        """ Deletes this Model's instance from the corresponding database table. If Model has been
        defined with soft-deletion (by defining corresponding attribute) - by default, the record
        from the database not deletes, but soft-deletion flag it set instead. This behavior may be
        changed by setting 'hard_delete' optional argument to True to override soft-deletion and
        actually delete corresponding record from the database table. """
        from ..request import OrmDeleteRequest
        if not self.__existing__:
            raise OrmInstanceMustBeWritten(
                "Cannot delete OrmModel's object instance because it is not written to the database yet!"
            )
        if self.__parent__:
            related = self.__parent__[0].__meta__.get_related_using_fk(self.__parent__[1])
            for c in related:
                if isinstance(c, RelatedToOneAbstract):
                    cv = self.__parent__[0].__getattrabs__(c.model_k)
                    cv.delete()
                    # cv.reset()
                    # cv.pretend_loaded()
                elif isinstance(c, OneToMany):
                    cv = getattr(self.__parent__[0], c.model_k)
                    cv.__detach__(self)
        if self.__m2m__:
            for m2m in set(self.__m2m__):
                m2m.delete(self)
        rq_session = session or self.__session__ or manager.get_session(instance=self)
        setattr(self, '__session__', rq_session)
        hard_deletion = self.__class__.__meta__.soft_delete is None or hard_delete
        self.__class__.__meta__.attrs_on_before_delete(self, hard_deletion)
        if self.__class__.__meta__.on_before_delete is not None \
                and self.__class__.__meta__.on_before_delete(self, hard_deletion) is False:
            return
        w = list()
        pk = self.__class__.__meta__.primary_key
        for k in pk:
            v = self.__getforwrite__(k, rq_session)
            c = getattr(self.__class__, k)
            w.append(c == v)
        OrmDeleteRequest(self.__class__, session=rq_session, hard_delete=hard_delete).where(*w).go()
        self.log.write_deletion()
        self.__class__.__meta__.attrs_on_after_delete(self, hard_deletion)
        if self.__class__.__meta__.on_after_delete is not None:
            self.__class__.__meta__.on_after_delete(self, hard_deletion)

    @classmethod
    def find(cls, what, **kwargs):
        """ Searching for given 'what' over records in the database of this Model. Searching will be
        done by selecting from the corresponding database table(s) filtering by Model's attributes which
        has been defined with 'findable=True' option. Not any attribute can be used in search query,
        depending on its (attribute's) type.
        @what:                  Textual phrase which search for;
        @returns:               Corresponding OrmRequest, which may be ran using 'go()' method or any
                                other corresponding one, or clarified before.
        """
        from ..request import OrmSearchRequest
        return OrmSearchRequest(cls, what, **kwargs)

    @property
    def __pk__(self):
        """ Returns a dict of attributes' keys (which constitude the primary key of this Model), and
        their corresponding values. """
        pk = self.__class__.__meta__.get_primary_key()
        r = dict()
        for c in pk:
            r[c.model_k] = getattr(self, c.model_k)
        return r


class OrmModelAliased(object):
    """ Overrider class used to name given Model with custom textual name in SELECTing request.
    The instance of this class returned by 'Model.as_()' method.
    """
    def __init__(self, model, alias):
        self.model = model
        self.alias = alias

    def __repr__(self):
        return "AlisedSource(%s AS '%s')" % (self.model.__name__, self.alias)


class CompositeOrmModel(OrmModelAbstract):
    """ Class which used to generate results of complex queries, when single Model cannot be used
    as source for results of the query. Usual case - when querying from the database only limited
    set of attributes (even from one Model) or joining other Models.
    """
    def describe(self, virtual=True, foreign=False, recursive=False):
        keys = self.__loadedkeys__
        p = list()
        p.append("<Complex result instance>")
        for k in keys:
            if k.startswith('__') and k.endswith('__'):
                continue
            c = self.__attributes__.get(k, None) or getattr(self.__class__, k, None)
            if isinstance(c, (list, tuple)):
                c = c[0]
            # if not c:
            #     continue
            # if not isinstance(c, AttributeAbstract):
            #     continue
            if isinstance(c, ForeignAttr) and not foreign:
                p.append("    %15s = (Foreign(%s.%s))" % (k, c.related_k, c.referenced_k))
                continue
            if isinstance(c, DataAttributeAbstract):
                if not virtual and isinstance(c, AttributeValueVirtual):
                    p.append("    %15s = (Virtual)" % k)
                    continue
                p.append("    %15s = \"%r\" (%s)" % (k, getattr(self, k), c.__class__.__name__))
                continue
            if isinstance(c, RelatedToOneAbstract):
                p.append("    %15s = (...ToOne relationship)" % k)
                continue
            if isinstance(c, RelatedToManyAbstract):
                p.append("    %15s = (...ToMany relationship)" % k)
                continue
            p.append("    %15s = %r" % (k, getattr(self, k)))
        return "\n".join(p)





from pc23 import xrange
from .base import KeyAttributeAbstract


class IndexKey(KeyAttributeAbstract):
    @property
    def key_name(self):
        return "%s_ix_%s" % (self.model.table, self.tbl_k)

    def declaration_sql(self, session):
        uk_name = self.tbl_k
        uk_keys = self.get_column_names(quoted=True, session=session)
        return "KEY %s (%s)" % (session.quote_name(uk_name), (",".join(uk_keys)))


class UniqueKey(KeyAttributeAbstract):
    @property
    def key_name(self):
        return "%s_uix_%s" % (self.model.table, self.tbl_k)

    def declaration_sql(self, session):
        uk_name = self.tbl_k
        uk_keys = self.get_column_names(quoted=True, session=session)
        return "CONSTRAINT %s UNIQUE (%s)" % (session.quote_name(uk_name), (",".join(uk_keys)))


def compare_keys(key1_columns, key2_columns):
    """ Compares two given simple (not foreign) keys by given _attrs. """
    if not isinstance(key1_columns, (list, tuple)):
        key1_columns = [key1_columns, ]
    if not isinstance(key2_columns, (list, tuple)):
        key2_columns = [key2_columns, ]
    sz1 = len(key1_columns)
    sz2 = len(key2_columns)
    if sz1 != sz2:
        return False
    for x in xrange(sz1):
        k1 = key1_columns[x]
        k2 = key2_columns[x]
        cn1 = k1 if isinstance(k1, str) else k1.tbl_k
        cn2 = k2 if isinstance(k2, str) else k2.tbl_k
        if cn1 != cn2:
            return False
    return True




from .base import FK_ONDELETE_CASCADE, FK_ONDELETE_SET_NULL, FK_ONDELETE_RESTRICT, FK_ONDELETE_ON_ACTION, Attribute, \
    Caption
from .derived import PASSWORD_ALGORYTHM_PLAIN, PASSWORD_ALGORYTHM_MD5, PASSWORD_ALGORYTHM_SHA1, \
    PASSWORD_ALGORYTHM_SHA224, PASSWORD_ALGORYTHM_SHA256, PASSWORD_ALGORYTHM_SHA384, PASSWORD_ALGORYTHM_SHA512, \
    Integer, BigInteger, PositiveInteger, PositiveBigInteger, Float, BigFloat, PositiveFloat, PositiveBigFloat, \
    Numeric, Currency, PositiveNumeric, String, Label, CaptionString, Uuid, Text, LongText, Boolean, Date, Time, \
    DateTime,Blob, LongBlob, UuidPrimaryKey, AutoIncrement, BigAutoIncrement, ObjectVersion, ObjectVersionTimestamp, \
    ObjectSoftdeleteAndVersion, SoftdeleteTimestamp, SoftdeleteBoolean, SoftdeleteDate, CreationTimestamp, Password, \
    YearMonth, Email, Phone, IMAddress, SocialNetworkAddress, IPAddress, IPMask, IPRequesties, Dimensions2D, \
    Dimensions3D, Box, Point, Vector, GeoCoordinate, IntegerChoice, StringChoice
from .virtual import Interval, IntInterval, FloatInterval, BigIntInterval, BigFloatInterval, DateInterval, \
    TimeInterval, DateTimeInterval, List, Dict, Matrix, FileAttachment, PictureAttachment, Files, Pictures, Set, \
    InlineSet
from .relationships import ForeignKey, VirtualForeignKey, Relationship, VirtualRelationship, OneToOne, ManyToOne, \
    OneToMany, ManyToMany, ForeignAttr, AuthorIdent, PerformerIdent, UserIdent
from .keys import IndexKey, UniqueKey
from .views import AttributeView, ViewJoin



from .base import AttributeAbstract, DataAttributeAbstract
from ..exceptions import OrmParameterError
from ..funcs import DbFunctionAbstract


class ViewPropertyAbstract(DataAttributeAbstract):
    pass


class AttributeView(ViewPropertyAbstract):
    def __init__(self, o, k=None, **kwargs):
        super(AttributeView, self).__init__(**kwargs)
        self.func = None
        if isinstance(o, DbFunctionAbstract):
            self.func = o
            o = None
            k = None
        elif k is None:
            if isinstance(o, AttributeAbstract):
                k = o.model_k
                o = o.model.__name__
            elif isinstance(o, str):
                if '.' in o:
                    if o.count('.') != 1:
                        raise OrmParameterError(
                            "AttributeView requires first argument to be an attribute, be a name of the attribute of"
                            " the base Model or be in format 'Model.attr_name'!"
                        )
                    o, k = o.split('.')
                else:
                    k = o
                    o = None
        self.owner = o
        self.owner_k = k
        self.kw = kwargs

    def on_view_metainit(self):
        if not self.owner:
            return
        base_model = getattr(self.model, '__model__')
        if base_model.__name__ == self.owner:
            base_c = getattr(base_model, self.owner_k)
        elif self.owner in self.model.__joins__:
            jc = getattr(self.model, self.owner)
            base_c = getattr(jc.ref_model, self.owner_k)
        else:
            return
        ck = dir(base_c)
        for k in ck:
            if k.startswith('__') and k.endswith('__'):
                continue
            if k in self.kw:
                continue
            v = getattr(base_c, k)
            if not hasattr(self, k) or not getattr(self, k, None):
                setattr(self, k, v)


class ViewJoin(ViewPropertyAbstract):
    def __init__(self, ref_model, *args):
        super(ViewJoin, self).__init__()
        self.ref_model = ref_model
        self.args = args





import inspect
from pc23 import xrange
from .mapper import mapper
from .exceptions import OrmException, OrmParameterError, OrmModelDeclarationError
from .request import OrmSelectRequest, OrmInsertRequest, OrmUpdateRequest, OrmDeleteRequest
from .results import OrmRawResult
from .manage import manager
from .attributes.base import ColumnAttributeAbstract
from .attributes.relationships import ManyToMany
from .values.base import AttributeValueAbstract
from .attributes.keys import compare_keys


# noinspection PyTypeChecker
class OrmSessionAbstract(object):
    """
    Abstract session class, must not be used in the end project.
    The session is the one connection to the DBMS within which requests are executed.
    """
    dbms_type = 'unknown'

    def __init__(self, host, dbname, username, password, port=None, tag=None, debug=None):
        """
        :host                       Database engine host name or IP;
        :dbname                     Database name on the database engine;
        :username                   Username which will be used to connect to the database;
        :password                   Corresponding password for the user above;
        :port                       Custom TCP port number if it is not the standard one;
        :tag                        Tag for this session (see documentation);
        :debug                      A set of debug flags, defaults to None;
        """
        self.dbms_host = host
        self.dbms_dbname = dbname
        self.dbms_username = username
        self.dbms_password = password
        self.dbms_port = port
        self.tag = tag
        self.cn = None
        self.cr = None
        self.time_zone = None
        self.dt_format = None
        self.session_user = False
        self.session_rls = False
        self.opened = False
        self.debug = debug if debug is not None else manager.get_debug()
        self.is_migrating = False
        self.open()

    def _check_opened(self):
        """ Ensures that session's connection been opened, raises an Exception otherwise. """
        if not self.opened:
            raise OrmException("The session is defined, but connection did not opened using `open` method!")

    @staticmethod
    def quote_name(value):  # Abstract
        pass

    def open(self):  # Abstract
        pass

    def close(self):
        """ Closes the previously opened connection to the DBMS. Doing nothing
        if connection did not established (method `open` did not called). """
        if not self.opened:
            return
        if self.cr:
            self.cr.close()
        if self.cn:
            self.cn.close()
        self.opened = False

    def set_time_zone(self, time_zone):
        """ Used to set_value the time zone on the DBMS side for this connection session.
        :time_zone                  (str) The time zone. Use DBMS correct known time zone.
                                    (None) If 'None' - reset DBMS time zone to the global default.
        """
        self._check_opened()
        self.time_zone = time_zone
        # TODO! Check PostgreSQL timezone syntax
        if self.time_zone is None:
            self.execute("SET time_zone = @@global.time_zone")
        else:
            self.execute("SET time_zone=%(time_zone)s", {'time_zone': self.time_zone})

    def set_session_user(self, user):
        """ Sets this session working user. The user here is not the database connection user used
        to deal with DBMS, but the representation of the information system's (which is using this
        ORM) user.
        :user                       (object instance) May be set to the corresponding Model's instance,
                                    (int|str) or may be set to the corresponding value of the primary
                                    key identifying the session user.
        """
        self.session_user = user

    def inqure_session_user(self):
        """
        :returns                    (None) if no session user is set;
                                    Value of the current session's user's primary key;
        """
        from .models.base import OrmModel
        if not self.session_user:
            return None
        if isinstance(self.session_user, OrmModel):
            _pk = self.session_user.__pk__.copy()
            pk = _pk.popitem()
            return pk[1]
        return self.session_user

    def begin(self):
        """ Begins transaction. `START TRANSACTION` using database library interface. """
        self._check_opened()
        self.cn.begin()

    def rollback(self):
        """ Rolls back the last transaction. `ROLLBACK` using database library interface. """
        self._check_opened()
        self.cn.rollback()

    def commit(self):
        """ Commits the transaction. `COMMIT` using database library interface. """
        self._check_opened()
        self.cn.commit()

    @staticmethod
    def _repr_query_params(p):
        """
        :returns                    Query parameters textual represented as Name=>Value
        """
        return ",  ".join([("%s=>%r" % (k, p[k])) for k in sorted(p.keys())])

    def execute(self, request, parameters=None):
        """ Executes the raw SQL query which not returning anything. Parameters in the
        query may be used using s(parameter_name)% format.
        :request                    Textual raw SQL query to execute to;
        :parameters                 [dict] with parameters for the query;
        """
        if not request:
            return
        if parameters is not None and not isinstance(parameters, dict):
            raise OrmParameterError("OrmSessionAbstract.query parameters must be set using dict() type!")
        if not manager.handle_query(request, parameters):
            print("DECLINED by handler: " + request)
            return
        rql = request.lower()
        do_debug = self.debug is True \
            or (isinstance(self.debug, (list, tuple))
                and (
                    ('select' in self.debug and (rql[:6] == 'select' or rql[:7] == '(select') and not self.is_migrating)
                    or ('migrate-getstruct' in self.debug and rql[:6] == 'select' and self.is_migrating)
                    or ('modify' in self.debug and rql[:6] in ('insert', 'update', 'delete'))
                    or ('alter' in self.debug and rql[:4] in ('crea', 'alte', 'drop'))
                ))
        if do_debug:
            print("")
            print(">> %s" % request)
        if parameters:
            p_ = dict()
            for k in parameters:
                v = parameters[k]
                v_ = v.value if isinstance(v, AttributeValueAbstract) else v
                p_[k] = v_
            parameters = p_
            if do_debug:
                print("@@ %s" % self._repr_query_params(parameters))
        self.cr.execute(request, parameters)
        if do_debug:
            print("")

    def query(self, request, parameters=None):
        """ Executes the raw SQL query and return `OrmRawResult` instance with results
        of the query. Parameters in the query may be used using s(parameter_name)% format.
        The results must be taken using '.all()' or '.one()' methods of the 'OrmRawResult'
        instance.
        :request                    Textual raw SQL query to execute to;
        :parameters                 [dict] with parameters for the query;
        :returns                    (OrmRawResult) results of the qurry.
        """
        self.execute(request, parameters)
        r = OrmRawResult(self.cr)
        # if self.debug is True or (isinstance(self.debug, str) and 'query-results' in self.debug):
        #     print("R<< %r" % r)
        #     print("|")
        return r

    def add(self, *args):
        """ Assigns specified object instance(s) with this session. """
        from .models.base import OrmModelAbstract
        for arg in args:
            if not isinstance(arg, OrmModelAbstract):
                raise OrmParameterError("`session`.add() method requires object instance(s) to be given!")
            arg.__session__ = self

    def new(self, model, *args, **kwargs):
        """ Creates a new object instance of given 'model' type and automatically assigns it with
        this session.
        :model                      (OrmModel) the model class instance of which to create;
        :*args, **kwargs            Optional parameters to pass to the creating instance;
        :returns                    (object instance) newly created object instance of given
                                    Model and already assigned with this session.
        """
        instance = model(*args, **kwargs)
        instance.__session__ = self
        return instance

    def select(self, *args, **kwargs):
        """ Selects from the database objects by given criteria.
        :returns                    (OrmSelectRequest) object ready to execute by .go() method.
        """
        from .models.views import OrmView
        kwargs['session'] = self
        _view = None
        _args = list()
        for arg in args:
            if not inspect.isclass(arg) or not issubclass(arg, OrmView):
                _args.append(arg)
                continue
            if _view is not None:
                raise OrmParameterError(
                    "can select from only one OrmView, several given!"
                )
            _view = arg
        if _view is not None:
            return _view.request(*_args, **kwargs)
        return OrmSelectRequest(*args, **kwargs)

    def insert(self, *args, **kwargs):
        """ Inserts given object instances in the database using this session and INSERT command.
        :returns                    (OrmInsertRequest) object ready to execute by .go() method.
        """
        kwargs['session'] = self
        return OrmInsertRequest(*args, **kwargs)

    def update(self, *args, **kwargs):
        """ Updates given object instances in the database using this session and UPDATE command.
        :returns                    (OrmUpdateRequest) object ready to execute by .go() method.
        """
        kwargs['session'] = self
        return OrmUpdateRequest(*args, **kwargs)

    def delete(self, *args, **kwargs):
        """ Deletes given object instances from the database using this session.
        :returns                    (OrmDeleteRequest) object ready to execute by .go() method.
        """
        kwargs['session'] = self
        return OrmDeleteRequest(*args, **kwargs)

    def write(self, *args, **kwargs):
        """ Writes given object instances to the database using this session. This method
        determines itself what to do - do INSERT or do UPDATE, so it is preferred against
        update() or insert() methods.
        """
        from .models.base import OrmModelAbstract
        kwargs['session'] = self
        for arg in args:
            if not isinstance(arg, OrmModelAbstract):
                raise OrmParameterError("session.write() method requires an object instance(s) to be given!")
            arg.write(**kwargs)

    def find(self, model, what, **kwargs):
        """
        Finds over objects for the given Model, returning found objects if any. Which attributes
        of the Model uses when looking for given phrase declares at Model declaration.
        :model                      (OrmModel) Model over objects of which type to search to;
        :what                       (str) Textual string what to find to;
        :returns                    (OrmSearchRequest) object ready to execute by .go() method.
        """
        from .request import OrmSearchRequest
        from .models.views import OrmView
        if issubclass(model, OrmView):
            return model.find(what, session=self, **kwargs)
        return OrmSearchRequest(model, what, session=self, **kwargs)

    def get_table_struct(self, table):  # Abstract
        pass

    def is_model_table_exists(self, model):  # Abstract
        pass

    def get_database_tables(self):  # Abstract
        pass

    def create_model_table(self, model, apply=True, restrict_new_fk=False):  # Abstract
        pass

    def alter_model_table(self, model, table_struct=None, drop=False, alter_foreign_keys=True):  # Abstract
        pass

    def create_model_foreign_keys(self, model):  # Abstract
        pass

    def drop_table_foreign_keys(self, tablename, tbl_struct_fks):  # Abstract
        pass

    def drop_model_table(self, model, apply=True):  # Abstract
        pass

    def migrate(self, drop=False, upload_initial=True):
        """ Migrates the database schema - all defined and written in the `mapper` real models - creating
        and modifying database at DBMS to confort to defined models.
        :drop                       [bool] Do drop columns which are not defined in altering Models;
        :upload_initial             [bool] Do upload initial data to the newly creating Models'
                                    tables, if any defined;
        """

        # Handling debugging rules
        self.is_migrating = True

        # Validating everything before continue (we must be sure that all elements are
        # correct and validated prior to database altering)
        mapper.validate_all()

        # Retrieving models for migration (excluding virtual models, like views and templates)
        models = mapper.get_models_for_migrate(self.tag)

        # Getting the list of existing tables
        existing_tables = self.get_database_tables()

        # Going through all defined models and filling up two lists - tables to be created
        # and tables to be altered.
        to_create = list()
        to_verify = list()
        to_drop = list()
        to_alter = dict()
        for name in models:
            model = models[name]
            if model.table in existing_tables:
                to_verify.append(name)
            else:
                to_create.append(name)

        # Stage 1. Retrieving structure of existing tables in the database by
        # declared models.
        tbl_struct = dict()
        for tablename in existing_tables:
            tbl_struct[tablename] = self.get_table_struct(tablename)

        # Stage 2. Analysing existing models to understand - do we need to
        # migrate something. We about to drop foreign keys prior to any
        # modifications because otherwise existing key may deny some actions
        # on existing tables.
        for name in to_verify:
            model = models[name]
            tablename = model.table
            queries = self.alter_model_table(model, tbl_struct[tablename], drop=drop, alter_foreign_keys=False)
            if not queries:
                continue
            to_alter[name] = queries

        # If no alter requests are present, and no new tables to be created
        # or existing tables to be dropped - returning.
        if not to_create and not to_alter and not to_drop:
            self.is_migrating = False
            return

        # Stage 3. If something is about to be done - dropping foreign keys
        # for affecting models first.
        for name in to_alter:
            model = models[name]
            tablename = model.table
            query = self.drop_table_foreign_keys(tablename, tbl_struct[tablename]['foreign_keys'])
            if not query:
                continue
            self.execute(query)

        # Stage 4. Create new tables, but not create foreign keys for them
        for name in to_create:
            model = models[name]
            self.create_model_table(model, apply=True, restrict_new_fk=True)

        # Stage 5. Uploading initial objects into newly created tables
        m2m_initial = dict()
        if upload_initial:
            for name in to_create:
                model = models[name]
                initial_objects = model.__meta__.initial
                if not initial_objects:
                    continue
                for d in initial_objects:
                    instance = model()
                    instance.__session__ = self
                    instance.__initialize_from__(**d)
                    instance.write(session=self, initial=True)
                    # Finding Many-To-Many attributes and adding data of them to
                    # initialize a little later
                    for k in d:
                        c = getattr(model, k, None)
                        if c is None or not isinstance(c, ManyToMany):
                            continue
                        objects_assigned = d[k]
                        if not isinstance(objects_assigned, (list, tuple)):
                            raise OrmModelDeclarationError(
                                "Many-To-Many assignments in __initial__ must be set using (list|tuple) type!"
                            )
                        c.validate()
                        if c.key not in m2m_initial:
                            m2m_initial[c.key] = {
                                'jmodel': c.junction_model,
                                'rows': []
                            }
                        f_model_name = c.key[0] if c.key[1] == model.__name__ else c.key[1]
                        l_columns = c.junction_model.__m2m_columns__[model.__name__]
                        f_columns = c.junction_model.__m2m_columns__[f_model_name]
                        f_keys = dict()
                        for c_ in f_columns:
                            f_keys[c_.parent_model_k] = c_.model_k
                        for oa in objects_assigned:
                            j_keys = dict()
                            for c_ in l_columns:
                                j_keys[c_.model_k] = instance.__getforwrite__(c_.parent_model_k, self)
                            if isinstance(oa, (int, str, float)):
                                if len(f_columns) != 1:
                                    raise OrmModelDeclarationError(
                                        "Many-To-Many assignment in __initial__ has simple format specified"
                                        " for the foreign object's primary key value, but foreign object's"
                                        " Model has complex primary key defined!"
                                    )
                                j_keys[f_columns[0].model_k] = oa
                            elif isinstance(oa, dict):
                                for oa_k in oa:
                                    if oa_k not in f_keys:
                                        raise OrmModelDeclarationError(
                                            "unknown Many-To-Many key '%s' in __initial__ for %s!"
                                            % (oa_k, str(c.key))
                                        )
                                    oa_v = oa[oa_k]
                                    j_keys[f_keys[oa_k]] = oa_v
                            else:
                                raise OrmModelDeclarationError(
                                    "Many-To-Many assignments in __initial__ must be set using dict "
                                    " of corresponding attributes' keys-values of foreign Model's primary"
                                    " key or using corresponding single value. (for example: {...'attrname':"
                                    " ['1', '2', '3'], ...} means that this object will be associated with"
                                    " foreign objects with primary key values 1, 2 and 3; or"
                                    " {...'attrname': [{'id': 1}, {'id': 2}, {'id: 3}], ...} - means the"
                                    " same)."
                                )
                            if j_keys in m2m_initial[c.key]['rows']:
                                continue
                            m2m_initial[c.key]['rows'].append(j_keys)

        # Stage 5a. Writting collected in the Stage 5 many to many relationships
        for m2m_key in m2m_initial:
            m2m = m2m_initial[m2m_key]
            m2m_jmodel = m2m['jmodel']
            m2m_rows = m2m['rows']
            for row in m2m_rows:
                jinstance = m2m_jmodel()
                for k_ in row:
                    setattr(jinstance, k_, row[k_])
                jinstance.write(session=self, all_attrs=True)

        # Stage 6. Alter existing tables, again, without any foreign key
        # affecting
        for name in to_alter:
            queries = to_alter[name]
            if queries is True:
                continue
            for query in queries:
                self.execute(query)

        # Stage 7. Create foreign keys for newly created tables
        for name in to_create:
            model = models[name]
            query = self.create_model_foreign_keys(model)
            if not query:
                continue
            self.execute(query)
        for name in to_alter:
            model = models[name]
            query = self.create_model_foreign_keys(model)
            if not query:
                continue
            self.execute(query)

        # Stage 8. Commiting changes (not necessary for MySQL, for example, but is
        # required for PostgreSQL, for example)
        self.commit()

        # Reseting debugging rules
        self.is_migrating = False


class MysqlSession(OrmSessionAbstract):
    dbms_type = 'mysql'

    @staticmethod
    def quote_name(value):
        return '`' + value + '`'

    def open(self):
        """ Opens the connection to the DBMS using stored at a time of class instance
        creation parameters like host, username, password and etc. """
        import MySQLdb
        self.cn = MySQLdb.connect(host=self.dbms_host,
                                  user=self.dbms_username,
                                  passwd=self.dbms_password,
                                  db=self.dbms_dbname,
                                  port=self.dbms_port or 3306,
                                  charset='utf8',
                                  use_unicode=True,
                                  client_flag=2)
        self.cr = self.cn.cursor()
        self.opened = True
        self.execute("SET NAMES 'utf8'")

    def get_table_struct(self, table):
        """ Reads the table structure from DBMS of the OrmModel. Reads attributes definitions and keys,
        including indexes, primary keys, foreign keys, unique keys, etc.
        :table                      (str) The existing database table name;
        :returns                    (dict) The existing database table structure.
        """
        keys = {}
        columns = {}
        indexes = {}
        foreign_keys = {}
        unique_keys = {}
        pk = []

        tablename = table if isinstance(table, str) else table.__table__

        # Selecting attributes definitions from INFORMATION_SCHEMA, basing on MySQL documentation
        # "22.4 The INFORMATION_SCHEMA COLUMNS Table"
        query = ("SELECT COLUMN_NAME, COLUMN_DEFAULT, IS_NULLABLE, DATA_TYPE, CHARACTER_MAXIMUM_LENGTH,"
                 " CHARACTER_OCTET_LENGTH, NUMERIC_PRECISION, NUMERIC_SCALE, DATETIME_PRECISION,"
                 " CHARACTER_SET_NAME, COLLATION_NAME, COLUMN_TYPE, COLUMN_KEY, EXTRA"
                 " FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_SCHEMA=%(ts)s AND TABLE_NAME=%(tn)s")
        dbr = self.query(query, {'ts': self.dbms_dbname, 'tn': tablename}).all()
        if not dbr:
            return None
        for r in dbr:
            c = dict()
            c_name = str(r[0])
            c['column_name'] = c_name
            c['column_default'] = r[1]
            c['is_nullable'] = True if str(r[2]).upper() == 'YES' else False
            c['data_type'] = str(r[3]).lower()
            c['character_maximum_length'] = r[4]
            c['character_octet_length'] = r[5]
            c['numeric_precision'] = r[6]
            c['numeric_scale'] = r[7]
            c['datetime_precision'] = r[8]
            c['character_set_name'] = str(r[9]).lower() if r[9] is not None else None
            c['collation_name'] = str(r[10]).lower() if r[10] is not None else None
            c['column_type'] = str(r[11]).lower()
            c['column_key'] = r[12]
            c['extra'] = str(r[13]).lower() if r[13] is not None else None
            columns[c_name] = c

        # Selecting informations about declared keys usage, basing on MySQL documentation
        # "22.11 The INFORMATION_SCHEMA KEY_COLUMN_USAGE Table"
        query = ("SELECT CONSTRAINT_NAME, COLUMN_NAME, ORDINAL_POSITION, POSITION_IN_UNIQUE_CONSTRAINT,"
                 " REFERENCED_TABLE_NAME, REFERENCED_COLUMN_NAME FROM INFORMATION_SCHEMA.KEY_COLUMN_USAGE"
                 " WHERE CONSTRAINT_SCHEMA = %(ts)s AND TABLE_NAME = %(tn)s"
                 " ORDER BY POSITION_IN_UNIQUE_CONSTRAINT, ORDINAL_POSITION")
        dbr = self.query(query, {'ts': self.dbms_dbname, 'tn': tablename}).all()
        if dbr:
            for r in dbr:
                key = dict()
                key_name = str(r[0])
                if key_name.upper() == 'PRIMARY' and r[1]:
                    # PRIMARY KEY part
                    pk.append(r[1])
                    continue
                if key_name in keys:
                    # Composite key part
                    keys[key_name]['_col_aliases'].append(r[1])
                    if r[5] is not None:
                        keys[key_name]['referenced_columns'].append(r[5])
                    continue
                # Defining a new key
                key['constraint_name'] = key_name
                # Basing on MySQL documentation: "22.11 The INFORMATION_SCHEMA KEY_COLUMN_USAGE Table"
                # the `POSITION_IN_UNIQUE_CONSTRAINT` column is NULL for UNIQUE KEYs and PRIMARY KEYs
                # and is not NULL for FOREIGN KEYs.
                key['is_foreign_key'] = r[3] is not None
                key['referenced_table_name'] = r[4]
                key['referenced_columns'] = [r[5], ] if r[5] is not None else []
                key['on_update'] = None
                key['on_delete'] = None
                key['_col_aliases'] = [r[1], ]
                keys[key_name] = key

        # Additionally selecting information about referential constraints - foreign keys. Doing it
        # by selecting from REFERENTIAL_CONSTRAINTS basing on MySQL documentation
        # "22.19 The INFORMATION_SCHEMA REFERENTIAL_CONSTRAINTS Table"
        query = ("SELECT CONSTRAINT_NAME, UPDATE_RULE, DELETE_RULE FROM"
                 " INFORMATION_SCHEMA.REFERENTIAL_CONSTRAINTS WHERE"
                 " CONSTRAINT_SCHEMA = %(ts)s AND TABLE_NAME = %(tn)s")
        dbr = self.query(query, {'ts': self.dbms_dbname, 'tn': tablename}).all()
        if dbr:
            for r in dbr:
                fk_name = r[0]
                if fk_name not in keys:
                    continue
                keys[fk_name]['is_foreign_key'] = True
                keys[fk_name]['on_update'] = r[1]
                keys[fk_name]['on_delete'] = r[2]

        # Handling got keys
        if keys:
            for name in keys:
                # We about to define three general types of keys:
                # a) PRIMARY KEY
                # b) FOREIGN KEY
                # c) UNIQUE KEY
                # INDEX KEY (or just a KEY in terms of MySQL) are stored
                # separatelly, so will not be found in general KEYS structure.
                # The first type (a) - the PRIMARY KEY (even composite)
                # detects in the procedure above and not stored in the general
                # `keys` variable, so this is not a point to work with PK here.
                key = keys[name]
                is_fk = key['is_foreign_key']
                if is_fk:
                    fk = {
                        'constraint_name': key['constraint_name'],
                        'model_idx_columns': key['_col_aliases'],
                        'ref_idx_columns': key['referenced_columns'],
                        'ref_table': key['referenced_table_name'],
                        'on_update': key['on_update'],
                        'on_delete': key['on_delete']
                    }
                    foreign_keys[name] = fk
                else:
                    uk = {
                        'constraint_name': key['constraint_name'],
                        '_col_aliases': key['_col_aliases']
                    }
                    unique_keys[name] = uk

        # Selecting information about INDEX KEYs, basing on MySQL documentation
        # "22.23 The INFORMATION_SCHEMA STATISTICS Table"
        query = ("SELECT INDEX_NAME, COLUMN_NAME, SUB_PART FROM INFORMATION_SCHEMA.STATISTICS"
                 " WHERE TABLE_SCHEMA = %(ts)s AND TABLE_NAME = %(tn)s")
        dbr = self.query(query, {'ts': self.dbms_dbname, 'tn': tablename}).all()
        if dbr:
            for r in dbr:
                name = r[0]
                # If the index key is already registered in the `keys` varitable - skipping
                # due this is not a index key.
                if name in keys or name.upper() == 'PRIMARY':
                    continue
                # If a composite index key part is found
                if name in indexes:
                    indexes[name]['_col_aliases'].append((r[1], r[2]))
                    indexes[name]['column_names'].append(r[1])
                    continue
                # Defining the new index key
                indexes[name] = {
                    'key_name': name,
                    '_col_aliases': [(r[1], r[2]), ],
                    'column_names': [r[1], ]
                }

        # Accumulating the resulting information about model's table definition
        table_struct = {
            'columns': columns,
            'indexes': indexes,
            'foreign_keys': foreign_keys,
            'unique_keys': unique_keys,
            'primary_key': pk
        }
        return table_struct

    def is_model_table_exists(self, model):
        """ Checks is the OrmModel's table existance in the database.
        :model                      (OrmModel) The Model for which to search for database table
                                    existabce to;
        :returns                    True if the corresponding database table is exists, False
                                    otherwise.
        """
        q = "SELECT TABLE_NAME FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_SCHEMA = %(ts)s AND TABLE_NAME = %(tn)s"
        r = self.query(q, {'tn': str(model.table), 'ts': self.dbms_dbname}).one()
        if not r:
            return False
        return str(r[0]) == str(model.table)

    def create_model_table(self, model, apply=True, restrict_new_fk=False):
        """ Creates a table in the database. Raises an exeption if the table is already
        exists in the database. Use `alter` instead to automatically migrate the database
        schema (do create or alter the table).
        :model                      (OrmModel) The corresponding Model for which database table to
                                    create to;
        :apply                      [bool] If set to True - automatically executes the ALTER query,
                                    if not - only returns the SQL query;
        :restrict_new_fk            [bool] If set to True - foreign keys will not be created right
                                    after database table creation; otherwise (default) - corresponding
                                    CREATE FOREIGN KEY procedures will be planned;
        :returns                    (str) Resulting SQL query.
        """
        if self.is_model_table_exists(model):
            raise Exception("Cannot CREATE TABLE because it is already exists: `%s`" % model.__table__)
        columns = model.__meta__.get_column_attributes()
        c_def = []
        for c in columns:
            q = c.declaration_sql(self)
            c_def.append("`%s` %s" % (c.tbl_k, q))
        q_def = ", ".join(c_def)
        if model.__meta__.primary_key:
            pkc = []
            for c in model.__meta__.get_primary_key():
                pkc.append("`%s`" % c.tbl_k)
            q_pk = "PRIMARY KEY (%s)" % (", ".join(pkc))
            q_def += ", " + q_pk
        foreign_keys = model.__meta__.get_foreign_keys()
        if foreign_keys and not restrict_new_fk:
            fks = []
            for fk in foreign_keys:
                fks.append(fk.declaration_sql(self))
            q_fk = ", ".join(fks)
            q_def += ", " + q_fk
        unique_keys = model.__meta__.get_unique_keys()
        if unique_keys:
            uks = []
            for uk in unique_keys:
                uks.append(uk.declaration_sql(self))
            q_uk = ", ".join(uks)
            q_def += ", " + q_uk
        index_keys = model.__meta__.get_index_keys()
        if index_keys:
            iks = []
            for ik in index_keys:
                iks.append(ik.declaration_sql(self))
            q_ik = ", ".join(iks)
            q_def += ", " + q_ik
        query = "CREATE TABLE `%s` (%s) ENGINE=InnoDB DEFAULT CHARSET=utf8 COLLATE=utf8_bin" % (model.table, q_def)
        if apply:
            self.execute(query)
        return query

    def alter_model_table(self, model, table_struct=None, drop=False, alter_foreign_keys=True):
        """ Returns a SQL query to alter database table basing on the declared Model.
        :model                      (OrmModel) the Model for which corresponding
                                    database table to alter;
        :table_struct               [dict] containing the current corresponding table structure
                                    at the database; if not set - current table structure will
                                    be automatically loaded by this method;
        :drop                       [bool] If set to True - columns in the database table which
                                    are not exists in the Model will be dropped; otherwise
                                    (default) - will be kept not touched;
        :alter_foreign_keys         [bool] If set to True (default) - foreign keys will be altered
                                    right in place of this alter procedure; otherwise this method
                                    will not worry about dropping and creating foreign keys;
        :returns                    (str) The resulting query.
        """
        if not table_struct:
            table_struct = self.get_table_struct(model.table)
        model_struct = model.__meta__.get_model_struct()

        _q_to_alter = []
        _fk_to_drop = []

        # Columns
        for k in model_struct['columns']:
            c = model_struct['columns'][k]
            tbl_k = c.tbl_k
            if not isinstance(c, ColumnAttributeAbstract):
                continue
            # If the column not exists at all - just add it in the alter procedure
            if tbl_k not in table_struct['columns']:
                add_sql = c.declaration_sql(self)
                q = "ADD `%s` %s" % (tbl_k, add_sql)
                _q_to_alter.append(q)
                continue
            # If the column exists - we have to compare it's model definition with
            # database table definition and make changes in the alter procedure
            # if some differences will be found.
            tbl_c = table_struct['columns'][tbl_k]
            if not c.is_eq_with_tbl_schema(tbl_c, self):
                alter_sql = c.declaration_sql(self)
                q = "CHANGE `%s` `%s` %s" % (tbl_k, tbl_k, alter_sql)
                _q_to_alter.append(q)
                # Checking - is this column using in foreign keys - and mark those
                # foreign keys where this column is using to drop prior to the
                # column been modified.
                for fk_name in table_struct['foreign_keys']:
                    if fk_name in _fk_to_drop:
                        continue
                    key = table_struct['foreign_keys'][fk_name]
                    if tbl_k not in key['model_idx_columns']:
                        continue
                    _fk_to_drop.append(fk_name)

        # Primary key
        t_pk = table_struct['primary_key']
        m_pk = model_struct['primary_key']
        if not compare_keys(t_pk, m_pk):
            cs = []
            for c in m_pk:
                cs.append("`%s`" % str(c.tbl_k))
            q = "" if not t_pk else "DROP PRIMARY KEY, "
            q += "ADD PRIMARY KEY(%s)" % ",".join(cs)
            _q_to_alter.append(q)

        # Unique keys
        t_uks = table_struct['unique_keys']
        m_uks = model_struct['unique_keys']
        keep_uks = []
        add_uks = []
        for k in m_uks:
            m_key = m_uks[k]
            if m_key.tbl_k not in t_uks:
                add_uks.append(m_key)
                continue
            t_key = t_uks[m_key.tbl_k]
            if not compare_keys(t_key['_col_aliases'], m_key.columns):
                add_uks.append(m_key)
            else:
                keep_uks.append(m_key.tbl_k)
        for tbl_k in t_uks:
            if tbl_k in keep_uks:
                continue
            q = "DROP KEY `%s`" % str(tbl_k)
            _q_to_alter.append(q)
        for c in add_uks:
            _q_to_alter.append("ADD %s" % c.declaration_sql(self))

        # Index keys
        t_ixs = table_struct['indexes']
        m_ixs = model_struct['indexes']
        keep_ixs = []
        add_ixs = []
        for k in m_ixs:
            m_key = m_ixs[k]
            if m_key.tbl_k not in t_ixs:
                add_ixs.append(m_key)
                continue
            t_key = t_ixs[m_key.tbl_k]
            if not compare_keys(t_key['column_names'], m_key.columns):
                add_ixs.append(m_key)
            else:
                keep_ixs.append(m_key.tbl_k)
        for tbl_k in t_ixs:
            if tbl_k in keep_ixs:
                continue
            q = "DROP KEY `%s`" % str(tbl_k)
            _q_to_alter.append(q)
        for c in add_ixs:
            _q_to_alter.append("ADD %s" % c.declaration_sql(self))

        # Foreign keys
        t_fks = table_struct['foreign_keys']
        m_fks = model_struct['foreign_keys']
        keep_fks = []
        add_fks = []
        for k in m_fks:
            m_key = m_fks[k]
            m_key.validate()
            if m_key.tbl_k not in t_fks or m_key.tbl_k in _fk_to_drop:
                add_fks.append(m_key)
                continue
            t_key = t_fks[m_key.tbl_k]
            if t_key['ref_table'] != m_key.ref_model.table \
                    or len(t_key['model_idx_columns']) != len(m_key.columns) \
                    or len(t_key['ref_idx_columns']) != len(m_key.ref_columns):
                add_fks.append(m_key)
                continue
            do_continue = False
            for i in xrange(len(m_key.columns)):
                if t_key['model_idx_columns'][i] == getattr(m_key.model, m_key.columns[i]).tbl_k:
                    continue
                add_fks.append(m_key)
                do_continue = True
                break
            if do_continue:
                continue
            for i in xrange(len(m_key.ref_columns)):
                if t_key['ref_idx_columns'][i] == getattr(m_key.ref_model, m_key.ref_columns[i]).tbl_k:
                    continue
                add_fks.append(m_key)
                do_continue = True
                break
            if do_continue:
                continue
            t_key_on_delete = str(t_key['on_delete']).upper()
            m_key_on_delete = 'RESTRICT' if m_key.on_delete is None else str(m_key.on_delete).upper()
            if t_key_on_delete != m_key_on_delete:
                add_fks.append(m_key)
                continue
            t_key_on_update = str(t_key['on_update']).upper()
            m_key_on_update = 'RESTRICT' if m_key.on_update is None else str(m_key.on_update).upper()
            if t_key_on_update != m_key_on_update:
                add_fks.append(m_key)
                continue
            keep_fks.append(m_key.tbl_k)
        for tbl_k in t_fks:
            if tbl_k in keep_fks or tbl_k in _fk_to_drop:
                continue
            _fk_to_drop.append(tbl_k)
        need_alter_foreign_keys = len(add_fks) != 0
        if alter_foreign_keys:
            for c in add_fks:
                _q_to_alter.append("ADD %s" % c.declaration_sql(self))
                _q_to_alter.append("ADD %s" % c.declaration_key_sql(self))

        # If nothing to do - just returning None
        if not _q_to_alter and not _fk_to_drop and not need_alter_foreign_keys:
            return None

        queries = list()
        # If we have foreign keys to be dropped prior to other ALTER actions,
        # including adding changed version of those foreign keys - put drops
        # in front of all other actions here
        if _fk_to_drop and alter_foreign_keys:
            q_fk_drop = []
            for fk_name in _fk_to_drop:
                q_fk_drop.append("DROP FOREIGN KEY `%s`" % str(fk_name))
                q_fk_drop.append("DROP KEY `%s`" % str(fk_name))
            dfk_query = "ALTER TABLE `%s` %s" % (model.table, ", ".join(q_fk_drop))
            queries.append(dfk_query)

        # Composing SQL ALTER request
        query = "ALTER TABLE `%s` %s" % (model.table, ", ".join(_q_to_alter)) if _q_to_alter else ""
        if query:
            queries.append(query)

        # If only we need - is to alter foreign keys
        if not queries and need_alter_foreign_keys:
            return True

        # Returning prepared queries
        return queries

    def create_model_foreign_keys(self, model):
        """ Returns a SQL query to create foreign keys for the database table
        basing on the declared Model.
        :model                      (OrmModel) the Model for which corresponding
                                    database table to create foreign keys for;
        :returns                    (str) The resulting query.
        """
        query = ""
        foreign_keys = model.__meta__.get_foreign_keys()
        if foreign_keys:
            fks = []
            for fk in foreign_keys:
                fks.append("ADD %s" % fk.declaration_sql(self))
                fks.append("ADD %s" % fk.declaration_key_sql(self))
            query = ", ".join(fks)
        if query:
            query = "ALTER TABLE `%s` %s" % (model.table, query)
        return query

    def drop_table_foreign_keys(self, tablename, tbl_struct_fks):
        """ Returns a SQL query to drop foreign keys for the database table.
        :tablename                  (str) The name of the datbase table;
        :tbl_struct_fks             (dict) Existing foreign keys struct;
        :returns                    (str) The resulting query.
        """
        if not tbl_struct_fks:
            return None
        queries = list()
        for fk_name in tbl_struct_fks:
            constraint_name = tbl_struct_fks[fk_name]['constraint_name']
            queries.append("DROP FOREIGN KEY `%s`" % str(constraint_name))
            queries.append("DROP KEY `%s`" % str(constraint_name))
        return "ALTER TABLE `%s` %s" % (tablename, ", ".join(queries))

    def drop_model_table(self, model, apply=True):
        """ Drops the model's table from the database.
        :apply          (bool) If set_value to True - automatically executes the ALTER query, if
                        not - only returns the SQL query.
        :returns        (str) SQL query for DROPing the table.
        """
        if not self.is_model_table_exists(model):
            raise Exception("Cannot DROP table because it is not exists: `%s`!" % model.__table__)
        pass

    def get_database_tables(self):
        """
        :returns                    (list) A list of table names in the current database.
        """
        q = "SELECT TABLE_NAME FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_SCHEMA = %(ts)s"
        dbr = self.query(q, {'ts': self.dbms_dbname}).all()
        existing_tables = []
        for r in dbr:
            existing_tables.append(r[0])
        return existing_tables


class PostgresqlSession(OrmSessionAbstract):
    """
    PostgreSQL session class
    """
    dbms_type = 'postgresql'

    def __init__(self, host, dbname, username, password, port=None, tag=None, schema='public', debug=None):
        self.dbms_schema = schema
        super(PostgresqlSession, self).__init__(host=host, dbname=dbname, username=username, password=password,
                                                port=port, tag=tag, debug=debug)

    @staticmethod
    def quote_name(value):
        return '"' + value + '"'

    def open(self):
        """ Opens the connection to the DBMS using stored at a time of class instance
        creation parameters like host, username, password and etc. """
        import psycopg2
        self.cn = psycopg2.connect(host=self.dbms_host,
                                   user=self.dbms_username,
                                   password=self.dbms_password,
                                   dbname=self.dbms_dbname,
                                   port=self.dbms_port or 5432)
        self.cr = self.cn.cursor()
        self.opened = True
        self.execute("SET NAMES 'utf8'")
        if self.dbms_schema != 'public':
            self.execute("SET search_path TO '%s'" % self.dbms_schema)

    def get_table_struct(self, table):
        """ Reads the table structure from DBMS of the OrmModel. Reads attributes definitions and keys,
        including indexes, primary keys, foreign keys, unique keys, etc.
        :table                      (str) The existing database table name;
        :returns                    (dict) The existing database table structure.
        """
        keys = {}
        columns = {}
        indexes = {}
        foreign_keys = {}
        unique_keys = {}
        pk = []
        pk_keynames = []

        tablename = table if isinstance(table, str) else table.__table__

        # Selecting attributes definitions from INFORMATION_SCHEMA, basing on PostgreSQL documentation
        # "34.15. The Information Schema.columns"
        query = ("SELECT COLUMN_NAME, COLUMN_DEFAULT, IS_NULLABLE, DATA_TYPE, CHARACTER_MAXIMUM_LENGTH,"
                 " CHARACTER_OCTET_LENGTH, NUMERIC_PRECISION, NUMERIC_SCALE, DATETIME_PRECISION, UDT_NAME"
                 " FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_SCHEMA=%(ts)s AND TABLE_NAME=%(tn)s")
        dbr = self.query(query, {'ts': self.dbms_schema, 'tn': tablename}).all()
        if not dbr:
            return None
        for r in dbr:
            c = dict()
            c_name = str(r[0])
            c['column_name'] = c_name
            c['column_default'] = r[1]
            c['is_nullable'] = True if str(r[2]).upper() == 'YES' else False
            c['data_type'] = str(r[3]).lower()
            c['character_maximum_length'] = r[4]
            c['character_octet_length'] = r[5]
            c['numeric_precision'] = r[6]
            c['numeric_scale'] = r[7]
            c['datetime_precision'] = r[8]
            c['character_set_name'] = None
            c['collation_name'] = None
            c['column_type'] = str(r[9]).lower()
            c['column_key'] = None
            c['extra'] = None
            columns[c_name] = c

        # Selecting informations about declared keys usage, basing on PostgreSQL documentation
        # "34.30 The Information Schema.key_column_usage"
        query = (
            "SELECT tc.CONSTRAINT_NAME, kcu.COLUMN_NAME, tc.CONSTRAINT_TYPE"
            " FROM INFORMATION_SCHEMA.TABLE_CONSTRAINTS tc"
            " JOIN INFORMATION_SCHEMA.KEY_COLUMN_USAGE kcu"
            " ON kcu.CONSTRAINT_NAME = tc.CONSTRAINT_NAME"
            " WHERE tc.CONSTRAINT_SCHEMA=%(ts)s AND tc.TABLE_NAME=%(tn)s"
            " AND UPPER(tc.CONSTRAINT_TYPE) IN ('PRIMARY KEY', 'FOREIGN KEY', 'UNIQUE')"
            " ORDER BY kcu.CONSTRAINT_NAME, kcu.POSITION_IN_UNIQUE_CONSTRAINT, kcu.ORDINAL_POSITION"
        )
        dbr = self.query(query, {'ts': self.dbms_schema, 'tn': tablename}).all()
        if dbr:
            for r in dbr:
                key = dict()
                key_name = str(r[0])
                key_type = str(r[2]).upper()
                if key_type == 'PRIMARY KEY' and r[1]:
                    # PRIMARY KEY part
                    pk.append(r[1])
                    pk_keynames.append(r[0])
                    continue
                if key_name in keys:
                    # Composite key part
                    keys[key_name]['_col_aliases'].append(r[1])
                    continue
                # Defining a new key
                key['constraint_name'] = key_name
                key['is_foreign_key'] = (key_type == 'FOREIGN KEY')
                key['referenced_table_name'] = None
                key['referenced_columns'] = []
                key['on_update'] = None
                key['on_delete'] = None
                key['_col_aliases'] = [r[1], ]
                keys[key_name] = key

        # Additionally selecting information about referential constraints - foreign keys. Doing it
        # by selecting from REFERENTIAL_CONSTRAINTS basing on PostgreSQL documentation
        # "34.32 The Information Schema.referential_constraints"
        # Note that PostgreSQL gives no posibility to filter referential_constraints table againts
        # specific table name and returns all constraints for the database at a time.
        query = ("SELECT CONSTRAINT_NAME, UPDATE_RULE, DELETE_RULE FROM"
                 " INFORMATION_SCHEMA.REFERENTIAL_CONSTRAINTS WHERE"
                 " CONSTRAINT_SCHEMA = %(ts)s")
        dbr = self.query(query, {'ts': self.dbms_schema}).all()
        if dbr:
            for r in dbr:
                fk_name = r[0]
                if fk_name not in keys:
                    continue
                keys[fk_name]['is_foreign_key'] = True
                keys[fk_name]['on_update'] = r[1]
                keys[fk_name]['on_delete'] = r[2]

        # Handling got keys
        if keys:
            for name in keys:
                # We about to define three general types of keys:
                # a) PRIMARY KEY
                # b) FOREIGN KEY
                # c) UNIQUE KEY
                # INDEX KEY (or just a KEY in terms of PostgreSQL) are stored
                # separatelly, so will not be found in general KEYS structure.
                # The first type (a) - the PRIMARY KEY (even composite)
                # detects in the procedure above and not stored in the general
                # `keys` variable, so this is not a point to work with PK here.
                key = keys[name]
                is_fk = key['is_foreign_key']
                if is_fk:
                    # Due PostgreSQL does not stores information about foreign key referenced
                    # table and columns in the information schema with key column usage - we have
                    # to query for this information additionaly againts every FK constraint.
                    query = (
                        "SELECT TABLE_NAME, COLUMN_NAME"
                        " FROM INFORMATION_SCHEMA.CONSTRAINT_COLUMN_USAGE"
                        " WHERE CONSTRAINT_NAME=%(cn)s"
                    )
                    dbr = self.query(query, {'cn': key['constraint_name']}).all()
                    if dbr:
                        ref_tablename = None
                        ref_columns = []
                        for r in dbr:
                            ref_tablename = str(r[0])
                            ref_columns.append(str(r[1]))
                        key['referenced_table_name'] = ref_tablename
                        key['referenced_columns'] = ref_columns
                    fk = {
                        'constraint_name': key['constraint_name'],
                        'model_idx_columns': key['_col_aliases'],
                        'ref_idx_columns': key['referenced_columns'],
                        'ref_table': key['referenced_table_name'],
                        'on_update': key['on_update'],
                        'on_delete': key['on_delete']
                    }
                    foreign_keys[name] = fk
                else:
                    uk = {
                        'constraint_name': key['constraint_name'],
                        '_col_aliases': key['_col_aliases']
                    }
                    unique_keys[name] = uk

        # Selecting information about INDEX KEYs, basing on internal PostgreSQL functionality.
        # Too bad, but this information is not accessible from standard schema (from
        # INFORMATION SCHEMA), so we have to use PostgreSQL specific way only.
        query = ("SELECT ix.relname, regexp_replace(pg_get_indexdef(indexrelid), '.*\\((.*)\\)', '\\1')"
                 " FROM pg_index i"
                 " JOIN pg_class t ON t.oid = i.indrelid"
                 " JOIN pg_class ix ON ix.oid = i.indexrelid"
                 " WHERE t.relname = %(tn)s")
        dbr = self.query(query, {'ts': self.dbms_schema, 'tn': tablename}).all()
        if dbr:
            for r in dbr:
                name = r[0]
                # If the index key is already registered in the `keys` varitable - skipping
                # due this is not a index key.
                if name in keys or name in pk_keynames:
                    continue
                indexes[name] = {
                    'key_name': name,
                    '_col_aliases': [],
                    'column_names': []
                }
                idx_columns = r[1].split(',')
                for cn in idx_columns:
                    indexes[name]['_col_aliases'].append((cn, None))
                    indexes[name]['column_names'].append(cn)

        # Accumulating the resulting information about model's table definition
        table_struct = {
            'columns': columns,
            'indexes': indexes,
            'foreign_keys': foreign_keys,
            'unique_keys': unique_keys,
            'primary_key': pk
        }
        return table_struct

    def is_model_table_exists(self, model):
        """ Checks is the OrmModel's table existance in the database.
        :model                      (OrmModel) The Model for which to search for database table
                                    existabce to;
        :returns                    True if the corresponding database table is exists, False
                                    otherwise.
        """
        q = "SELECT TABLE_NAME FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_SCHEMA = %(ts)s AND TABLE_NAME = %(tn)s"
        r = self.query(q, {'tn': self.dbms_schema, 'ts': self.dbms_dbname}).one()
        if not r:
            return False
        return str(r[0]) == str(model.table)

    def create_model_table(self, model, apply=True, restrict_new_fk=False):
        """ Creates a table in the database. Raises an exeption if the table is already
        exists in the database. Use `alter` instead to automatically migrate the database
        schema (do create or alter the table).
        :model                      (OrmModel) The corresponding Model for which database table to
                                    create to;
        :apply                      [bool] If set to True - automatically executes the ALTER query,
                                    if not - only returns the SQL query;
        :restrict_new_fk            [bool] If set to True - foreign keys will not be created right
                                    after database table creation; otherwise (default) - corresponding
                                    CREATE FOREIGN KEY procedures will be planned;
        :returns                    (str) Resulting SQL query.
        """
        if self.is_model_table_exists(model):
            raise Exception("Cannot CREATE TABLE because it is already exists: \"%s\"" % model.table)
        columns = model.__meta__.get_column_attributes()
        c_def = []
        for c in columns:
            q = c.declaration_sql(self)
            c_def.append("\"%s\" %s" % (c.tbl_k, q))
        q_def = ", ".join(c_def)
        if model.__meta__.primary_key:
            pkc = []
            for c in model.__meta__.get_primary_key():
                pkc.append("\"%s\"" % c.tbl_k)
            q_pk = "PRIMARY KEY (%s)" % (", ".join(pkc))
            q_def += ", " + q_pk
        foreign_keys = model.__meta__.get_foreign_keys()
        if foreign_keys and not restrict_new_fk:
            fks = []
            for fk in foreign_keys:
                fks.append(fk.declaration_sql(self))
            q_fk = ", ".join(fks)
            q_def += ", " + q_fk
        unique_keys = model.__meta__.get_unique_keys()
        if unique_keys:
            uks = []
            for uk in unique_keys:
                uks.append(uk.declaration_sql(self))
            q_uk = ", ".join(uks)
            q_def += ", " + q_uk
        #
        # PostgreSQL cannot create or alter indexes within table definition or modification,
        # it uses separate 'CREATE INDEX', 'ALTER INDEX' AND 'DROP INDEX' procedures, so them must
        # be handled separatelly, not within main query.
        # index_keys = model.__meta__.get_index_keys()
        #
        # if index_keys:
        #     iks = []
        #     for ik in index_keys:
        #         iks.append(ik.declaration_sql(self))
        #     q_ik = ", ".join(iks)
        #     q_def += ", " + q_ik
        query = "CREATE TABLE \"%s\" (%s)" % (model.table, q_def)
        if apply:
            self.execute(query)
        return query

    def alter_model_table(self, model, table_struct=None, drop=False, alter_foreign_keys=True):
        """ Returns a SQL query to alter database table basing on the declared Model.
        :model                      (OrmModel) the Model for which corresponding
                                    database table to alter;
        :table_struct               [dict] containing the current corresponding table structure
                                    at the database; if not set - current table structure will
                                    be automatically loaded by this method;
        :drop                       [bool] If set to True - columns in the database table which
                                    are not exists in the Model will be dropped; otherwise
                                    (default) - will be kept not touched;
        :alter_foreign_keys         [bool] If set to True (default) - foreign keys will be altered
                                    right in place of this alter procedure; otherwise this method
                                    will not worry about dropping and creating foreign keys;
        :returns                    (str) The resulting query.
        """
        if not table_struct:
            table_struct = self.get_table_struct(model.table)
        model_struct = model.__meta__.get_model_struct()

        _q_to_alter = []
        _fk_to_drop = []

        # Columns
        for k in model_struct['columns']:
            c = model_struct['columns'][k]
            tbl_k = c.tbl_k
            if not isinstance(c, ColumnAttributeAbstract):
                continue
            # If the column not exists at all - just add it in the alter procedure
            if tbl_k not in table_struct['columns']:
                add_sql = c.declaration_sql(self)
                q = "ADD \"%s\" %s" % (tbl_k, add_sql)
                _q_to_alter.append(q)
                continue
            # If the column exists - we have to compare it's model definition with
            # database table definition and make changes in the alter procedure
            # if some differences will be found.
            tbl_c = table_struct['columns'][tbl_k]
            if not c.is_eq_with_tbl_schema(tbl_c, self):
                alter_sql = c.alter_sql(self)
                if isinstance(alter_sql, (list, tuple)):
                    for q_ in alter_sql:
                        _q_to_alter.append("ALTER COLUMN \"%s\" %s" % (tbl_k, q_))
                else:
                    q = "ALTER COLUMN \"%s\" %s" % (tbl_k, alter_sql)
                    _q_to_alter.append(q)
                # Checking - is this column using in foreign keys - and mark those
                # foreign keys where this column is using to drop prior to the
                # column been modified.
                for fk_name in table_struct['foreign_keys']:
                    if fk_name in _fk_to_drop:
                        continue
                    key = table_struct['foreign_keys'][fk_name]
                    if tbl_k not in key['model_idx_columns']:
                        continue
                    _fk_to_drop.append(fk_name)

        # Primary key
        t_pk = table_struct['primary_key']
        m_pk = model_struct['primary_key']
        if not compare_keys(t_pk, m_pk):
            cs = []
            for c in m_pk:
                cs.append("\"%s\"" % str(c.tbl_k))
            q = "" if not t_pk else "DROP PRIMARY KEY, "
            q += "ADD PRIMARY KEY(%s)" % ",".join(cs)
            _q_to_alter.append(q)

        # Unique keys
        t_uks = table_struct['unique_keys']
        m_uks = model_struct['unique_keys']
        keep_uks = []
        add_uks = []
        for k in m_uks:
            m_key = m_uks[k]
            if m_key.tbl_k not in t_uks:
                add_uks.append(m_key)
                continue
            t_key = t_uks[m_key.tbl_k]
            if not compare_keys(t_key['_col_aliases'], m_key.columns):
                add_uks.append(m_key)
            else:
                keep_uks.append(m_key.tbl_k)
        for tbl_k in t_uks:
            if tbl_k in keep_uks:
                continue
            q = "DROP CONSTRAINT \"%s\"" % str(tbl_k)
            _q_to_alter.append(q)
        for c in add_uks:
            _q_to_alter.append("ADD %s" % c.declaration_sql(self))

        # Index keys
        #
        # PostgreSQL cannot create or alter indexes within table definition or modification,
        # it uses separate 'CREATE INDEX', 'ALTER INDEX' AND 'DROP INDEX' procedures, so them must
        # be handled separatelly, not within main query.
        # index_keys = model.__meta__.get_index_keys()
        #
        # t_ixs = table_struct['indexes']
        # m_ixs = model_struct['indexes']
        # keep_ixs = []
        # add_ixs = []
        # for k in m_ixs:
        #     m_key = m_ixs[k]
        #     if m_key.tbl_k not in t_ixs:
        #         add_ixs.append(m_key)
        #         continue
        #     t_key = t_ixs[m_key.tbl_k]
        #     if not compare_keys(t_key['column_names'], m_key.columns):
        #         add_ixs.append(m_key)
        #     else:
        #         keep_ixs.append(m_key.tbl_k)
        # for tbl_k in t_ixs:
        #     if tbl_k in keep_ixs:
        #         continue
        #     q = "DROP KEY `%s`" % str(tbl_k)
        #     _q_to_alter.append(q)
        # for c in add_ixs:
        #     _q_to_alter.append("ADD %s" % c.declaration_sql(self))

        # Foreign keys
        t_fks = table_struct['foreign_keys']
        m_fks = model_struct['foreign_keys']
        keep_fks = []
        add_fks = []
        for k in m_fks:
            m_key = m_fks[k]
            m_key.validate()
            if m_key.tbl_k not in t_fks or m_key.tbl_k in _fk_to_drop:
                add_fks.append(m_key)
                continue
            t_key = t_fks[m_key.tbl_k]
            if t_key['ref_table'] != m_key.ref_model.table \
                    or len(t_key['model_idx_columns']) != len(m_key.columns) \
                    or len(t_key['ref_idx_columns']) != len(m_key.ref_columns):
                add_fks.append(m_key)
                continue
            do_continue = False
            for i in xrange(len(m_key.columns)):
                if t_key['model_idx_columns'][i] == getattr(m_key.model, m_key.columns[i]).tbl_k:
                    continue
                add_fks.append(m_key)
                do_continue = True
                break
            if do_continue:
                continue
            for i in xrange(len(m_key.ref_columns)):
                if t_key['ref_idx_columns'][i] == getattr(m_key.ref_model, m_key.ref_columns[i]).tbl_k:
                    continue
                add_fks.append(m_key)
                do_continue = True
                break
            if do_continue:
                continue
            t_key_on_delete = str(t_key['on_delete']).upper()
            if t_key_on_delete in ('NO ACTION', 'RESTRICT'):
                t_key_on_delete = 'RESTRICT'
            m_key_on_delete = 'RESTRICT' if m_key.on_delete is None else str(m_key.on_delete).upper()
            if t_key_on_delete != m_key_on_delete:
                add_fks.append(m_key)
                continue
            t_key_on_update = str(t_key['on_update']).upper()
            if t_key_on_update in ('NO ACTION', 'RESTRICT'):
                t_key_on_update = 'RESTRICT'
            m_key_on_update = 'RESTRICT' if m_key.on_update is None else str(m_key.on_update).upper()
            if t_key_on_update != m_key_on_update:
                add_fks.append(m_key)
                continue
            keep_fks.append(m_key.tbl_k)
        for tbl_k in t_fks:
            if tbl_k in keep_fks or tbl_k in _fk_to_drop:
                continue
            _fk_to_drop.append(tbl_k)
        need_alter_foreign_keys = len(add_fks) != 0
        if alter_foreign_keys:
            for c in add_fks:
                _q_to_alter.append("ADD %s" % c.declaration_sql(self))

        # If nothing to do - just returning None
        if not _q_to_alter and not _fk_to_drop:
            return None

        queries = list()
        # If we have foreign keys to be dropped prior to other ALTER actions,
        # including adding changed version of those foreign keys - put drops
        # in front of all other actions here
        if _fk_to_drop and alter_foreign_keys:
            q_fk_drop = []
            for fk_name in _fk_to_drop:
                q_fk_drop.append("DROP CONSTRAINT \"%s\"" % str(fk_name))
            dfk_query = "ALTER TABLE \"%s\" %s" % (model.table, ", ".join(q_fk_drop))
            queries.append(dfk_query)

        # Composing SQL ALTER request
        query = "ALTER TABLE \"%s\" %s" % (model.table, ", ".join(_q_to_alter)) if _q_to_alter else ""
        if query:
            queries.append(query)

        # If only we need - is to alter foreign keys
        if not queries and need_alter_foreign_keys:
            return True

        # Returning prepared queries
        return queries

    def create_model_foreign_keys(self, model):
        """ Returns a SQL query to create foreign keys for the database table
        basing on the declared Model.
        :model                      (OrmModel) the Model for which corresponding
                                    database table to create foreign keys for;
        :returns                    (str) The resulting query.
        """
        query = ""
        foreign_keys = model.__meta__.get_foreign_keys()
        if foreign_keys:
            fks = []
            for fk in foreign_keys:
                fks.append("ADD %s" % fk.declaration_sql(self))
            query = ", ".join(fks)
        if query:
            query = "ALTER TABLE \"%s\" %s" % (model.table, query)
        return query

    def drop_table_foreign_keys(self, tablename, tbl_struct_fks):
        """ Returns a SQL query to drop foreign keys for the database table.
        :tablename                  (str) The name of the datbase table;
        :tbl_struct_fks             (dict) Existing foreign keys struct;
        :returns                    (str) The resulting query.
        """
        if not tbl_struct_fks:
            return None
        queries = list()
        for fk_name in tbl_struct_fks:
            constraint_name = tbl_struct_fks[fk_name]['constraint_name']
            queries.append("DROP CONSTRAINT \"%s\"" % str(constraint_name))
        return "ALTER TABLE \"%s\" %s" % (tablename, ", ".join(queries))

    def drop_model_table(self, model, apply=True):
        """ Drops the model's table from the database.
        :apply          (bool) If set_value to True - automatically executes the ALTER query, if
                        not - only returns the SQL query.
        :returns        (str) SQL query for DROPing the table.
        """
        if not self.is_model_table_exists(model):
            raise Exception("Cannot DROP table because it is not exists: `%s`!" % model.__table__)
        pass

    def get_database_tables(self):
        """
        :returns                    (list) A list of table names in the current database.
        """
        q = "SELECT TABLE_NAME FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_SCHEMA = '%s'" % self.dbms_schema
        dbr = self.query(q).all()
        existing_tables = []
        for r in dbr:
            existing_tables.append(r[0])
        return existing_tables




from pc23 import xrange
from ..exceptions import OrmAttributeIsProtected, OrmModelClassMismatch
from ..mapper import mapper
from ..manage import manager
from .base import AttributeValueVirtual


class ValueRelatedToOne(AttributeValueVirtual):
    def delete(self):
        sz = len(self.c.foreign_c.columns)
        triggers_enabled = self.instance.__triggers__
        for x in xrange(sz):
            local_k = self.c.foreign_c.columns[x]
            setattr(self.instance, local_k, None)
        self.instance.__triggers__ = triggers_enabled
        self.reset()
        self.pretend_loaded()

    def set_value(self, instance, key, value):
        from ..results import OrmRelatedObjectNotAssigned
        self.c.validate()
        self.c.foreign_c.validate()
        if value is not None and not isinstance(value, self.c.foreign_model):
            raise OrmModelClassMismatch(
                "assigning object instance of type %s while required %s!"
                % (type(value), self.c.foreign_model.__name__)
            )
        if not getattr(value, '__existing__', True):
            value.write(session=self.instance.__session__)
        sz = len(self.c.foreign_c.columns)
        triggers_enabled = instance.__triggers__
        for x in xrange(sz):
            local_k = self.c.foreign_c.columns[x]
            if not value:
                setattr(instance, local_k, None)
                continue
            ref_k = self.c.foreign_c.ref_columns[x]
            ref_value = getattr(value, ref_k)
            setattr(instance, local_k, ref_value)
        instance.__triggers__ = triggers_enabled
        self._loaded = True
        self._value = value if value else OrmRelatedObjectNotAssigned(instance, self.c.foreign_c)

    def load(self):
        from ..results import OrmRelatedObjectNotAssigned
        self._loaded = True
        self.c.validate()
        foreign_model = self.c.foreign_model
        foreign_c = self.c.foreign_c
        session = self.instance.__session__
        if session is None:
            return OrmRelatedObjectNotAssigned(self.instance, foreign_c)
        foreign_c.validate()
        w = list()
        notnone = False
        for x in xrange(len(foreign_c.columns)):
            if foreign_c.model == self.instance.__class__:
                fc_ = getattr(foreign_model, foreign_c.ref_columns[x])
                lv_ = getattr(self.instance, foreign_c.columns[x])
            else:
                fc_ = getattr(foreign_model, foreign_c.columns[x])
                lv_ = getattr(self.instance, foreign_c.ref_columns[x])
            if lv_:
                notnone = True
            w.append(fc_ == lv_)
        child = session.select(foreign_model).where(*w).limit(1).go().first if notnone else None
        if child is None:
            return OrmRelatedObjectNotAssigned(self.instance, foreign_c)
        child.__setattrabs__('__parent__', (self.instance, foreign_c))
        return child

    def pretend_loaded(self):
        from ..results import OrmRelatedObjectNotAssigned
        super(ValueRelatedToOne, self).pretend_loaded()
        self.c.validate()
        foreign_c = self.c.foreign_c
        self._value = OrmRelatedObjectNotAssigned(self.instance, foreign_c)


class ValueRelatedToMany(AttributeValueVirtual):
    def set_value(self, instance, key, value):
        raise OrmAttributeIsProtected("the ...ToMany relationship attribute cannot be assigned to the value!")

    def load(self):
        from ..results import OrmResultOneToMany, OrmResult
        self._loaded = True
        self.c.validate()
        foreign_model = self.c.foreign_model
        foreign_c = self.c.foreign_c
        session = self.instance.__session__
        if session is None:
            r = OrmResultOneToMany(OrmResult())
            r.model = foreign_model
            r.relation_c = foreign_c
            r.parent_instance = self.instance
            return r
        foreign_c.validate()
        w = list()
        for x in xrange(len(foreign_c.columns)):
            if foreign_c.model == self.instance.__class__:
                fc_ = getattr(foreign_model, foreign_c.ref_columns[x])
                lv_ = getattr(self.instance, foreign_c.columns[x])
            else:
                fc_ = getattr(foreign_model, foreign_c.columns[x])
                lv_ = getattr(self.instance, foreign_c.ref_columns[x])
            w.append(fc_ == lv_)
        childs = session.select(foreign_model).where(*w).go()
        r = OrmResultOneToMany(childs)
        r.model = foreign_model
        r.relation_c = foreign_c
        r.parent_instance = self.instance
        for instance in r.objects:
            instance.__setattrabs__('__parent__', (self.instance, foreign_c))
        return r


class ValueManyToMany(AttributeValueVirtual):
    def __init__(self, session=None):
        super(ValueManyToMany, self).__init__()
        self.session = session

    def set_value(self, instance, key, value):
        raise OrmAttributeIsProtected("the ManyToMany relationship attribute cannot be assigned to the value!")

    def load(self):
        from ..session import OrmSessionAbstract
        from ..results import OrmResultManyToMany

        self._loaded = True
        self.c.validate()

        if not self.instance.__existing__:
            # TODO! If instance not been written yet - do an empty list!
            pass

        self.session = manager.get_session(session=self.session, instance=self.instance)
        if not isinstance(self.session, OrmSessionAbstract):
            # TODO! If instance not been written yet - do an empty list!
            pass

        local_model = self.c.model
        local_model_name = self.c.key[0] if self.c.key[0] == self.c.model.__name__ else self.c.key[1]
        ref_model_name = self.c.key[1] if self.c.key[0] == self.c.model.__name__ else self.c.key[0]
        ref_model = mapper.get_model_by_name(ref_model_name)
        join_columns = self.c.junction_model.__m2m_columns__[ref_model.__name__]
        c_columns = self.c.junction_model.__m2m_columns__[local_model.__name__]
        j_args = list()
        for jc in join_columns:
            mc = getattr(ref_model, jc.parent_model_k)
            j_args.append(jc == mc)
        w = list()
        for jc in c_columns:
            kv = self.instance.__getattrabs__(jc.parent_model_k)
            jjc = getattr(self.c.junction_model, jc.model_k)
            w.append(jjc == kv)
        dbr = self.session.select(ref_model).join(self.c.junction_model, *j_args).where(*w).get(ref_model).go()
        r = OrmResultManyToMany(dbr)
        r.parent_instance = self.instance
        r.relation_c = self.c
        r.local_model = local_model
        r.ref_model = ref_model
        r.local_model_name = local_model_name
        r.ref_model_name = ref_model_name
        r.junction_model = self.c.junction_model
        r.m2m_key = self.c.key
        return r





"""

GenORM
~~~~~~

Simple functional MySQL/PostgreSQL ORM.

Author:         Denis Khodus aka "Reagent"
License:        MIT

"""

from .manage import manager as orm
from .attributes import *
from .values import *
from .models import *
from .funcs import *
from .session import MysqlSession, PostgresqlSession
from .request import OrmSelectRequest, OrmUpdateRequest, OrmInsertRequest, OrmDeleteRequest, OrmUndeleteRequest
from .results import OrmRawResult, OrmResult
from .exceptions import *


__version__ = "0.1.5"





import datetime
from .base import AttributeValueVirtual, AttributeValueNull
from ..exceptions import OrmValueError


class ValueInterval(AttributeValueVirtual):
    def __init__(self):
        super(ValueInterval, self).__init__()

    def __repr__(self):
        return "[%r,%r]" % (self.low, self.high)

    def _get_c(self):
        return getattr(self.instance.__class__, self.model_k) if self.model_k is not None else None

    def _check_value(self, value):
        c = self._get_c()
        childs_type = c.childs_type
        if value is None or isinstance(value, AttributeValueNull):
            if c is not None and not c.nullable:
                raise OrmValueError("cannot be NULL because this interval attribute is not nullable!")
            return
        if (isinstance(value, int) and childs_type not in ('int', 'float')) \
                or (isinstance(value, float) and childs_type != 'float') \
                or (isinstance(value, datetime.datetime) and childs_type != 'datetime') \
                or (isinstance(value, datetime.date) and childs_type not in ('date', 'datetime')) \
                or (isinstance(value, datetime.time) and childs_type not in ('time', 'datetime')):
            raise OrmValueError("interval values must be '%s' type!" % childs_type)
        elif not isinstance(value, (int, float, datetime.date, datetime.time, datetime.datetime)):
            raise OrmValueError("intervals can handle only numeric and date/time values!")

    @property
    def nullable(self):
        return self._get_c().nullable if self._get_c() is not None else True

    @property
    def min(self):
        return self._get_c().min_value if self._get_c() is not None else None

    @property
    def max(self):
        return self._get_c().max_value if self._get_c() is not None else None

    @property
    def low(self):
        return getattr(self.instance, self._get_c().k_low) if self._get_c() is not None else None

    @low.setter
    def low(self, value):
        self._check_value(value)
        if self.min is not None and value < self.min:
            value = self.min
        setattr(self.instance, self._get_c().k_low, value)

    @property
    def high(self):
        return getattr(self.instance, self._get_c().k_high) if self._get_c() is not None else None

    @high.setter
    def high(self, value):
        self._check_value(value)
        if self.max is not None and value > self.max:
            value = self.max
        setattr(self.instance, self._get_c().k_high, value)

    @property
    def range(self):
        return self.low, self.high

    @range.setter
    def range(self, value):
        if isinstance(value, ValueInterval):
            self.low, self.high = value.low, value.high
            return
        if value is None or isinstance(value, AttributeValueNull):
            c = self._get_c()
            if c is not None and not c.nullable:
                raise OrmValueError("cannot be NULL because this interval attribute is not nullable!")
            self.low = None
            self.high = None
            return
        if not isinstance(value, (tuple, list)) or len(value) != 2:
            raise OrmValueError("interval must be set using its low|high properties or (low,high) list/tuple!")
        self.low = value[0]
        self.high = value[1]

    def get_value(self):
        return self

    def set_value(self, instance, key, value):
        self.range = value
        return self.low, self.high




from pc23 import xrange
from collections import OrderedDict
from .exceptions import OrmModelClassMismatch, OrmParameterError, OrmInstanceNotFound
from .utils import gettext


class OrmRawResult(object):
    """ Special class which generates by raw query and having the results DBMS returned """
    def __init__(self, cr):
        self.cr = cr
        self._result = self.cr.fetchall()

    def all(self):
        """Return all got from DBMS rows as list of rows of _attrs ([[],[],[]], [[],[],[]]...] """
        return self._result

    def one(self):
        """Return only first row got from DBMS as list of _attrs ([],[],[]...) """
        return self._result[0] if self._result else None

    def __len__(self):
        """ Return quantity of rows in the result """
        return len(self._result)


class OrmResult(tuple):
    def __init__(self):
        super(OrmResult, self).__init__()
        self._total_qty = None
        self.objects = list()
        self.session = None
        self.request = None
        self.page = None
        self.per_page = None
        self.query = None
        self.params = None
        self.rows = None
        self._initialized = True

    def __getitem__(self, item):
        return self.objects[item]

    def __contains__(self, item):
        return item in self.objects

    def __iter__(self):
        for x in self.objects:
            yield x

    def __len__(self):
        return len(self.objects)

    def __repr__(self):
        return self.objects.__repr__()

    def __reversed__(self):
        return self.objects.__reversed__()

    def write(self, session=None, force=False):
        if not self.objects:
            return self
        for instance in self.objects:
            instance.write(session=session, force=force)
        return self

    @property
    def total_qty(self):
        if self._total_qty:
            return self._total_qty
        self._total_qty = self.request.count()
        return self._total_qty

    def one(self, x=None):
        return None if not self.objects else (self.objects[0] if x is None else self.objects[x])

    @property
    def first(self):
        return self.one(0)

    @property
    def last(self):
        return self.one(-1)

    @property
    def sole(self):
        return self.first

    def set(self, **kwargs):
        for instance in self.objects:
            for k in kwargs:
                setattr(instance, k, kwargs[k])
        return self

    def delete(self):
        for instance in self.objects:
            instance.delete()
        return self

    def describe(self):
        if not self.objects:
            return "<None objects>"
        r = list()
        for instance in self.objects:
            r.append(instance.describe())
        return "\n\n-----\n\n".join(r)

    # def __setattr__(self, key, value):
    #     if key.startswith('__') and key.endswith('__'):
    #         super(OrmResult, self).__setattr__(key, value)
    #         return
    #     if key in dir(OrmResult) or key in dir(self):
    #         super(OrmResult, self).__setattr__(key, value)
    #         return
    #     if not getattr(self, '_initialized', False):
    #         super(OrmResult, self).__setattr__(key, value)
    #         return
    #     for instance in self.objects:
    #         setattr(instance, key, value)


class OrmRelatedObjectNotAssigned(object):
    def __init__(self, parent_instance, relation_c):
        self.__parent_instance__ = parent_instance
        self.__relation_c__ = relation_c
        self.__parent_k__ = None

    def __nonzero__(self):
        return False

    def __bool__(self):
        return False

    def __getattribute__(self, key):
        if key in dir(OrmRelatedObjectNotAssigned):
            return super(OrmRelatedObjectNotAssigned, self).__getattribute__(key)
        if key in dir(self):
            return super(OrmRelatedObjectNotAssigned, self).__getattribute__(key)
        child = self.create()
        return child.__getattribute__(key)

    def __setattr__(self, key, value):
        if key in ('__parent_instance__', '__relation_c__', '__parent_k__'):
            super(OrmRelatedObjectNotAssigned, self).__setattr__(key, value)
            return
        child = self.create()
        child.__setattr__(key, value)

    def create(self):
        rel_c = self.__relation_c__
        model = rel_c.ref_model if rel_c.model == self.__parent_instance__.__class__ else rel_c.model
        child = model()
        return self.assign_to(child)

    def assign_to(self, child):
        child.__setattrabs__('__parent__', (self.__parent_instance__, self.__relation_c__))
        setattr(self.__parent_instance__, self.__parent_k__, child)
        return child


class OrmResultOneToMany(OrmResult):
    def __init__(self, src):
        super(OrmResultOneToMany, self).__init__()
        self.objects = list(src.objects)
        self.session = src.session
        self.request = src.request
        self.page = src.page
        self.per_page = src.per_page
        self.query = src.query
        self.params = src.params
        self.rows = src.rows
        self.base_model = None
        self.parent_instance = None
        self.relation_c = None
        self.model = None
        self._to_delete = list()

    def __detach__(self, x):
        y = None
        if isinstance(x, int):
            if x > len(self.objects):
                raise IndexError("object instance's index out of range")
            y = x
        else:
            for i in xrange(len(self.objects)):
                if self.objects[i] != x:
                    continue
                y = i
                break
        if y is None:
            return
        del(self.objects[y])

    def _flag_modified(self):
        pass

    def add(self, x=None, **kwargs):
        from .models.meta import OrmModelMeta
        from .models.base import OrmModelAbstract
        from .attributes.relationships import RelatedToManyAbstract
        if x and isinstance(x, (OrmModelMeta, OrmModelAbstract)):
            if x.__class__ != self.model.__class__:
                raise OrmModelClassMismatch(
                    "Appending object instance has class `%s` while requred `%s`!" %
                    (x.__class__.__name__, self.model.__class__.__name__)
                )
            self.objects.append(x)
            x.__setattrabs__('__parent__', (self.parent_instance, self.relation_c))
            self._flag_modified()
            return
        instance = self.model()
        if kwargs:
            for k in kwargs:
                value = kwargs[k]
                c = getattr(self.model, k)
                if isinstance(value, (list, tuple)) and isinstance(c, RelatedToManyAbstract):
                    childs = getattr(instance, k)
                    for ci in value:
                        if not isinstance(ci, dict):
                            raise OrmParameterError(
                                "childs values must be a type of list of dict!"
                            )
                        childs.add(**ci)
                    continue
                setattr(instance, k, value)
        self.objects.append(instance)
        instance.__setattrabs__('__parent__', (self.parent_instance, self.relation_c))
        self._flag_modified()

    def remove(self, x):
        if x > len(self.objects):
            raise IndexError("object instance's index out of range")
        instance = self.objects[x]
        self._to_delete.append(instance)
        del(self.objects[x])
        self._flag_modified()

    def reverse(self):
        self.objects.reverse()

    def sort(self, attribute, reverse=False, textual=True):
        # TODO!
        if not self.objects:
            return

    def write(self, session=None, force=False):
        for instance in self._to_delete:
            instance.delete()
        self._to_delete = list()
        super(OrmResultOneToMany, self).write(session=session, force=force)


class OrmResultManyToMany(OrmResult):
    def __init__(self, src):
        super(OrmResultManyToMany, self).__init__()
        self.objects = list(src.objects)
        self.session = src.session
        self.request = src.request
        self.page = src.page
        self.per_page = src.per_page
        self.query = src.query
        self.params = src.params
        self.rows = src.rows
        self.base_model = None
        self.parent_instance = None
        self.relation_c = None
        self.local_model = None
        self.ref_model = None
        self.local_model_name = None
        self.ref_model_name = None
        self.junction_model = None
        self.m2m_key = None
        self._to_delete = list()
        self._local_pks = None
        self._assigned = dict()
        self._init_assigned()

    def _init_assigned(self):
        for instance in self.objects:
            pks = self._get_foreign_pks(instance)
            self._assigned[instance] = pks
            self._set_instance_m2m(instance)

    def _find_by_instance(self, f_instance):
        foreign_pks = self._get_foreign_pks(f_instance)
        return self._find_by_pks(foreign_pks)

    def _find_by_pks(self, foreign_pks):
        for fi in self._assigned:
            fpks = self._get_foreign_pks(fi)
            if fpks == foreign_pks:
                return fi
        return None

    def _get_local_pks(self):
        if self._local_pks is not None:
            return self._local_pks
        pks = OrderedDict()
        for c in self.parent_instance.__meta__.get_primary_key():
            k = c.model_k
            v = self.parent_instance.__getattrabs__(k)
            pks[k] = v
        self._local_pks = pks
        return pks

    @staticmethod
    def _get_foreign_pks(f_instance):
        pks = OrderedDict()
        for c in f_instance.__class__.__meta__.get_primary_key():
            k = c.model_k
            v = f_instance.__getattrabs__(k)
            pks[k] = v
        return pks

    def _get_index_of_instance(self, instance):
        return self.objects.index(instance) if instance in self.objects else None

    def _set_instance_m2m(self, instance):
        instance.__m2m__.add(self)

    def _drop_instance_m2m(self, instance):
        instance.__m2m__.discard(self)

    def _assign(self, instance):
        if self._find_by_instance(instance):
            self._set_instance_m2m(instance)
            return
        self.objects.append(instance)
        self._set_instance_m2m(instance)

    def _remove(self, x):
        from .models.meta import OrmModelMeta
        from .models.base import OrmModelAbstract
        if isinstance(x, int):
            if x > len(self.objects):
                raise IndexError("out of range in Many-To-Many relationship for index %i!" % x)
            instance = self.objects[x]
        elif isinstance(x, dict):
            instance = self._find_by_pks(x)
            if not instance:
                raise IndexError("instance with given primary key values is not assigned with this Many-To-Many!")
        elif isinstance(x, (OrmModelMeta, OrmModelAbstract)):
            instance = x
        else:
            raise OrmParameterError(
                "Many-To-Many requires list index, instance or primary key values to be given for deletion!"
            )
        x = self._get_index_of_instance(instance)
        if x is None:
            raise IndexError("given assigned object instance is not found for this Many-To-Many!")
        del(self.objects[x])
        self._drop_instance_m2m(instance)

    def _flag_modified(self):
        self.parent_instance.__set_modified__(self.relation_c.model_k, gettext('Changed'), None)

    def is_assigned(self, f_instance):
        """ Verifies is the given instance already being assigned. """
        return self._find_by_instance(f_instance) is not None

    def find(self, *args, **kwargs):
        """ Tries to find given instance in the list of assigned instance objects by instance
        or by primary key value. Return found object instance if found, None otherwise. """
        if args and len(args) > 1:
            raise OrmParameterError("`ManyToMany`.`find` method requires only ONE instance to be given to find!")
        instance = args[0] if args else None
        if instance is not None:
            return self._find_by_instance(instance)
        elif kwargs:
            return self._find_by_pks(kwargs)
        else:
            raise OrmParameterError(
                "`ManyToMany`.`find` method requires an instance or a primary key values to be given to"
                " find in the junction!"
            )

    def add(self, *args, **kwargs):
        if args and kwargs:
            raise OrmParameterError(
                "`ManyToMany`.`add` requires a foreign instance(s) OR primary key value to be given, but both"
                " instance(s) and primary key value specified instead!"
            )
        elif not args and not kwargs:
            raise OrmParameterError(
                "`ManyToMany`.`add` requires a foreign instance(s) or primary key value to be given, but none"
                " is given!"
            )
        if args:
            for x in args:
                if not isinstance(x, self.ref_model):
                    raise OrmParameterError(
                        "Appending object instance has class `%s` while requred `%s`!" %
                        (x.__class__.__name__, self.ref_model.__class__.__name__)
                    )
                self._assign(x)
        else:
            instance = self._find_by_pks(kwargs)
            if instance:
                return
            w = list()
            for k in kwargs:
                if not hasattr(self.ref_model, k):
                    raise OrmParameterError("OrmModel `%s` has no attribute named `%s`!" % (self.ref_model_name, k))
                w.append(getattr(self.ref_model, k) == kwargs[k])
            r = self.session.select(self.ref_model).where(*w).limit(1).src(self.ref_model).go().one()
            if not r:
                raise OrmInstanceNotFound("Instance with given conditions has not been found!")
            self._assign(r)
        self._flag_modified()

    def delete(self, *args, **kwargs):
        if not args and not kwargs:
            raise OrmParameterError(
                "`ManyToMany`.`delete` requires object instanc(es) OR list index(es) OR primary key values to be"
                " given, but none of them been specified!"
            )
        elif args and kwargs:
            raise OrmParameterError(
                "`ManyToMany`.`delete` requires object instanc(es) OR list index(es) OR primary key values to be"
                " given, but not all at a time!"
            )
        if args:
            for x in args:
                self._remove(x)
        else:
            self._remove(kwargs)
        self._flag_modified()

    def write(self, session=None, force=False):
        session = session or self.session
        w = list()
        m2m_local = self.junction_model.__m2m_columns__[self.local_model_name]
        m2m_ref = self.junction_model.__m2m_columns__[self.ref_model_name]
        for c in m2m_local:
            parent_k = c.parent_model_k
            parent_value = self.parent_instance.__getforwrite__(parent_k, session)
            w.append(c == parent_value)
        existing = session.select(self.junction_model).where(*w).go()
        for fi in self.objects:
            ref_v = dict()
            do_insert = True
            for c in m2m_ref:
                ref_v[c.model_k] = fi.__getforwrite__(c.parent_model_k, session)
            for ei in existing:
                eq = True
                for k in ref_v:
                    v = ref_v[k]
                    if v != ei.__getforwrite__(k, session):
                        eq = False
                        break
                if eq:
                    do_insert = False
                    break
            if not do_insert:
                continue
            ji = self.junction_model()
            for c in m2m_local:
                setattr(ji, c.model_k, self.parent_instance.__getforwrite__(c.parent_model_k, session))
            for c in m2m_ref:
                setattr(ji, c.model_k, fi.__getforwrite__(c.parent_model_k, session))
            ji.__setattrabs__('__session__', session)
            ji.write()
        for ei in existing:
            ref_v = dict()
            avoid_delete = False
            for c in m2m_ref:
                ref_v[c.parent_model_k] = ei.__getforwrite__(c.model_k, session)
            for fi in self.objects:
                eq = True
                for k in ref_v:
                    v = ref_v[k]
                    if v != fi.__getforwrite__(k, session):
                        eq = False
                        break
                if eq:
                    avoid_delete = True
                    break
            if avoid_delete:
                continue
            ei.delete()


class OrmModelUpdateAction(object):
    def __init__(self, model, session, values):
        self._model = model
        self._session = session
        self._values = values
        self._where = None

    def where(self, *args):
        from .funcs import where_
        self._where = where_(args)
        return self

    def go(self):
        objects = self._model.get(session=self._session) \
            if not self._where \
            else self._model.get(self._where, session=self._session)
        if not objects:
            return
        for obj in objects:
            for k in self._values:
                if not hasattr(obj, k):
                    continue
                setattr(obj, k, self._values[k])
            obj.write()




from collections import OrderedDict
from ..attributes.base import AttributeAbstract, ColumnAttributeAbstract, KeyAttributeAbstract,\
    VirtualAttributeAbstract, DataAttributeAbstract
from ..values import AttributeValueVirtual
from ..attributes.relationships import RelatedAbstract, ManyToMany, Relationship, VirtualRelationship, ForeignAttr, \
    RelatedToOneAbstract, ForeignKeyConstraintAbstract, OneToMany, FK_ONDELETE_SET_NULL
from ..exceptions import OrmParameterError


class OrmModelMetaStruct(object):
    def __init__(self, cls):
        self.PROTECTED = ('cls', 'auto_increment', 'primary_key')
        self.tag = None
        self.session = None
        self.cls = cls
        self.tablename = None
        self.prefix = None
        self.primary_key = None
        self.auto_increment = None
        self.userident = False
        self.soft_delete = None
        self.default_order = None
        self.version_key = None
        self.caption_key = None
        self.value_key = None
        self.initial = None
        self.on_before_write = None
        self.on_after_write = None
        self.on_insert = None
        self.on_update = None
        self.on_change = None
        self.on_before_delete = None
        self.on_after_delete = None
        self.on_instance_create = None
        self.on_instance_loaded = None
        self.constructor = None
        self.log = None
        self.foreign = False
        self.foreign_keys = list()
        self.unique_keys = list()
        self.index_keys = list()
        self.relationships = list()
        self.author_ident = list()
        self.performer_ident = list()
        self.timestamp_creation = list()
        self.timestamp_modification = list()

    def get_model_struct(self):
        """ Returns the structure of this OrmModel, which describes defined attributes:
        attributes, keys (including foreign keys), primary key and etc. This method is using
        at a moment of database migration mechanism, for example. """
        return {
            'columns': self.get_column_attributes_dict(),
            'indexes': self.get_index_keys_dict(),
            'foreign_keys': self.get_foreign_keys_dict(),
            'unique_keys': self.get_unique_keys_dict(),
            'primary_key': self.get_primary_key()
        }

    def get_primary_key(self):
        """ Returns primary key (list of attributes) of the Model. """
        pk = []
        for k in self.primary_key:
            pk.append(getattr(self.cls, k))
        return pk

    def get_autoincrement(self):
        """ Returns auto-increment attribute if one defined, None otherwise. """
        return getattr(self.cls, self.auto_increment) if self.auto_increment else None

    def get_attributes(self):
        """ Returns a list of attributes of the Model (attributes themself, not keys). """
        attrs_ = list()
        keys = dir(self.cls)
        for k in keys:
            if k.startswith('__') and k.endswith('__'):
                continue
            c = type.__getattribute__(self.cls, k)
            if not isinstance(c, AttributeAbstract):
                continue
            attrs_.append((c.declaration_order, c))
        if not attrs_:
            return []
        return [y[1] for y in sorted(attrs_, key=lambda x: x[0])]

    def get_attributes_keys(self):
        """ Returns a list of attributes' names (keys) of the Model. """
        attrs_ = list()
        keys = dir(self.cls)
        for k in keys:
            if k.startswith('__') and k.endswith('__'):
                continue
            c = type.__getattribute__(self.cls, k)
            if not isinstance(c, AttributeAbstract):
                continue
            attrs_.append((c.declaration_order, k))
        if not attrs_:
            return []
        return [y[1] for y in sorted(attrs_, key=lambda x: x[0])]

    def get_real_attributes(self):
        """ Returns a list of attributes which has direct reflection in the database table
        (attributes themself, not keys). """
        attrs = list()
        keys = dir(self.cls)
        for k in keys:
            if k.startswith('__') and k.endswith('__'):
                continue
            c = type.__getattribute__(self.cls, k)
            if not isinstance(c, (ColumnAttributeAbstract, KeyAttributeAbstract)):
                continue
            attrs.append(c)
        return attrs

    def get_data_attributes(self):
        """ Returns a list of data attributes (attributes which holding some kind of data, attributes
        themselfs, not keys). """
        attrs = list()
        keys = dir(self.cls)
        for k in keys:
            if k.startswith('__') and k.endswith('__'):
                continue
            c = type.__getattribute__(self.cls, k)
            if not isinstance(c, DataAttributeAbstract):
                continue
            attrs.append(c)
        return attrs

    def get_findable_attributes(self):
        """ Returns a list of findable attributes, which has property 'findable' set to True -- only
        'real' attributes which are reflecting to the database table (attributes themself, not keys). """
        attrs = list()
        keys = dir(self.cls)
        for k in keys:
            if k.startswith('__') and k.endswith('__'):
                continue
            c = type.__getattribute__(self.cls, k)
            if not isinstance(c, ColumnAttributeAbstract):
                continue
            if not c.findable:
                continue
            attrs.append(c)
        return attrs

    def get_related_attributes(self):
        """ Returns a list of relationships of any X-to-X type , attributes themself, not keys. """
        attrs_ = list()
        keys = dir(self.cls)
        for k in keys:
            if k.startswith('__') and k.endswith('__'):
                continue
            c = type.__getattribute__(self.cls, k)
            if not isinstance(c, RelatedAbstract):
                continue
            attrs_.append((c.declaration_order, c))
        return [y[1] for y in sorted(attrs_, key=lambda x: x[0])]

    def get_related_to_one_attributes(self):
        """ Returns a list of relationships of type ...ToOne (OneToOne and ManyToOne), attributes
        themself, not keys. """
        attrs_ = list()
        keys = dir(self.cls)
        for k in keys:
            if k.startswith('__') and k.endswith('__'):
                continue
            c = type.__getattribute__(self.cls, k)
            if not isinstance(c, RelatedToOneAbstract):
                continue
            attrs_.append((c.declaration_order, c))
        return [y[1] for y in sorted(attrs_, key=lambda x: x[0])]

    def get_related_using_fk(self, fk):
        """ Returns a list of relationship attributes (..ToOne) which using given ForeignKey,
        attributes themself, not keys. """
        if isinstance(fk, str):
            fkc = getattr(self.cls, fk, None)
            if fkc is None:
                raise OrmParameterError(
                    "get_related_using_fk() requires that 'fk' argument be a type of (str) -- the name of"
                    " the corresponding ForeignKey attribute, or ForeignKey attribute itself!"
                )
        elif not isinstance(fk, ForeignKeyConstraintAbstract):
            raise OrmParameterError(
                "get_related_using_fk() requires that 'fk' argument be a type of (str) -- the name of"
                " the corresponding ForeignKey attribute, or ForeignKey attribute itself!"
            )
        else:
            fkc = fk
        attrs_ = list()
        rattrs = self.get_related_attributes()
        for c in rattrs:
            c.validate()
            if c.foreign_c != fkc:
                continue
            attrs_.append((c.declaration_order, c))
        return [y[1] for y in sorted(attrs_, key=lambda x: x[0])]

    def get_attributes_for_select(self):
        """ Returns a list of attributes used in selecing request (returns 'real' attributes, themself,
        not keys, which are directly reflected in database table and has corresponding single columns). """
        from ..funcs import DbFunctionAbstract
        attrs_ = list()
        keys = dir(self.cls)
        for k in keys:
            if k.startswith('__') and k.endswith('__'):
                continue
            c = type.__getattribute__(self.cls, k)
            if not isinstance(c, (DataAttributeAbstract, DbFunctionAbstract)):
                continue
            if isinstance(c, ForeignAttr):
                continue
            if isinstance(c, DataAttributeAbstract) and c.virtual:
                continue
            attrs_.append((c.declaration_order, c))
        return [y[1] for y in sorted(attrs_, key=lambda x: x[0])]

    def get_attributes_for_write(self):
        """ Returns a list of attributes used in modify requests (UPDATE and INSERT) -- 'real'
        attributes which has direct reflection in database table (attributes themself, not keys). """
        attrs_ = list()
        keys = dir(self.cls)
        for k in keys:
            if k.startswith('__') and k.endswith('__'):
                continue
            c = type.__getattribute__(self.cls, k)
            if not isinstance(c, ColumnAttributeAbstract):
                continue
            attrs_.append((c.declaration_order, c))
        return [y[1] for y in sorted(attrs_, key=lambda x: x[0])]

    def get_keys_for_write(self):
        """ Returns a list of attributes' keys used in modify requests (UPDATE and INSERT) -- 'real'
        attributes which has direct reflection in database table (Keys only). """
        attrs_ = list()
        keys = dir(self.cls)
        for k in keys:
            if k.startswith('__') and k.endswith('__'):
                continue
            c = type.__getattribute__(self.cls, k)
            if not isinstance(c, ColumnAttributeAbstract):
                continue
            attrs_.append((c.declaration_order, k))
        return [y[1] for y in sorted(attrs_, key=lambda x: x[0])]

    def get_virtualkeys_for_write(self, instance):
        """ Returns a list of virtual attributes' keys which has routines for writting and
        been already loaded at a time of Model instance writting. """
        attrs_ = list()
        keys = dir(self.cls)
        for k in keys:
            if k.startswith('__') and k.endswith('__'):
                continue
            c = type.__getattribute__(self.cls, k)
            if not isinstance(c, VirtualAttributeAbstract):
                continue
            v = instance.__getattrabs__(k)
            is_loaded = getattr(v, '_loaded', True)
            if not is_loaded:
                continue
            attrs_.append((c.declaration_order, k))
        return [y[1] for y in sorted(attrs_, key=lambda x: x[0])]

    def get_related_loaded_attributes(self, instance):
        """ Returns a list of relationship attributes (X-To-X) which been loaded
        from the database (initially some attributes not loaded from database -- for
        example, ManyToMany and OneToMany, and loads only when trying to read them
        in the program code, saving some server resources). Relationships which
        not been loaded not fall into this list. """
        attrs_ = list()
        keys = dir(self.cls)
        for k in keys:
            if k.startswith('__') and k.endswith('__'):
                continue
            c = type.__getattribute__(self.cls, k)
            if not isinstance(c, (RelatedAbstract, ManyToMany)):
                continue
            v = instance.__getattrabs__(k)
            if isinstance(v, AttributeValueVirtual):
                continue
            attrs_.append((c.declaration_order, c))
        return [y[1] for y in sorted(attrs_, key=lambda x: x[0])]

    def get_related_loaded_keys(self, instance):
        """ Returns a list of relationship attributes' keys (X-To-X) which been loaded
        from the database (initially some attributes not loaded from database -- for
        example, ManyToMany and OneToMany, and loads only when trying to read them
        in the program code, saving some server resources). Relationships which
        not been loaded not fall into this list. """
        attrs_ = list()
        keys = dir(self.cls)
        for k in keys:
            if k.startswith('__') and k.endswith('__'):
                continue
            c = type.__getattribute__(self.cls, k)
            if not isinstance(c, (RelatedAbstract, ManyToMany)):
                continue
            v = instance.__getattrabs__(k)
            if not isinstance(v, AttributeValueVirtual):
                continue
            if not v.loaded:
                continue
            attrs_.append((c.declaration_order, k))
        return [y[1] for y in sorted(attrs_, key=lambda x: x[0])]

    def get_foreign_attrs(self):
        attrs_ = list()
        keys = dir(self.cls)
        for k in keys:
            if k.startswith('__') and k.endswith('__'):
                continue
            c = type.__getattribute__(self.cls, k)
            if not isinstance(c, ForeignAttr):
                continue
            attrs_.append((c.declaration_order, c))
        return [y[1] for y in sorted(attrs_, key=lambda x: x[0])]

    def get_foreign_attrs_keys(self):
        attrs_ = list()
        keys = dir(self.cls)
        for k in keys:
            if k.startswith('__') and k.endswith('__'):
                continue
            c = type.__getattribute__(self.cls, k)
            if not isinstance(c, ForeignAttr):
                continue
            attrs_.append((c.declaration_order, k))
        return [y[1] for y in sorted(attrs_, key=lambda x: x[0])]

    def get_foreign_attrs_by_relationships(self):
        foreign_attrs = self.get_foreign_attrs()
        rel_cc = dict()
        for c in foreign_attrs:
            c.validate()
            related_k = c.related_k
            rel_c = getattr(self.cls, related_k)
            rel_c.validate()
            if rel_c not in rel_cc:
                rel_cc[rel_c] = list()
            rel_cc[rel_c].append(c)
        return rel_cc

    def get_relationships(self):
        cc = list()
        for k in self.relationships:
            cc.append(getattr(self.cls, k))
        return cc

    def get_column_attributes(self):
        attrs = dir(self.cls)
        raw_c = list()
        for k in attrs:
            if k.startswith('__') and k.endswith('__'):
                continue
            c = getattr(self.cls, k)
            if not isinstance(c, ColumnAttributeAbstract):
                continue
            raw_c.append(c)
        return sorted(raw_c, key=lambda c_: getattr(c_, 'declaration_order', 65535))

    def get_column_attributes_dict(self):
        cc = self.get_column_attributes()
        dc = OrderedDict()
        for c in cc:
            dc[c.model_k] = c
        return dc

    def get_foreign_keys(self):
        cc = list()
        for k in self.foreign_keys:
            c = getattr(self.cls, k)
            c.validate()
            cc.append(c)
        return cc

    def get_foreign_keys_dict(self):
        cc = self.get_foreign_keys()
        dc = dict()
        for c in cc:
            dc[c.model_k] = c
        return dc

    def get_unique_keys(self):
        cc = list()
        for k in self.unique_keys:
            c = getattr(self.cls, k)
            c.validate()
            cc.append(c)
        return cc

    def get_unique_keys_dict(self):
        cc = self.get_unique_keys()
        dc = dict()
        for c in cc:
            dc[c.model_k] = c
        return dc

    def get_index_keys(self):
        cc = list()
        for k in self.index_keys:
            c = getattr(self.cls, k)
            c.validate()
            cc.append(c)
        return cc

    def get_index_keys_dict(self):
        cc = self.get_index_keys()
        dc = dict()
        for c in cc:
            dc[c.model_k] = c
        return dc

    def get_relationship_with_model(self, model):
        if not self.relationships:
            return None
        result = None
        for k in self.relationships:
            c = getattr(self.cls, k)
            c.validate()
            if c.ref_model == model:
                if result is not None:
                    return True
                result = c
                break
        return result

    def make_relationship_with_model(self, model, virtual=False, on_delete=None):
        ref_pk = model.__meta__.get_primary_key()
        if len(ref_pk) > 1:
            raise OrmParameterError(
                "OrmModel '%s' has complex primary key and cannot be automatically mapped to the"
                " OrmModel '%s' using relationship!" % (model.__name__, self.cls.__name__)
            )
        pk = ref_pk[0]
        pk_k = pk.model_k
        attr_cls = Relationship if not virtual else VirtualRelationship
        on_delete = on_delete or FK_ONDELETE_SET_NULL
        c = attr_cls(model, on_delete=on_delete)
        k = "%s_%s" % (model.__name__.lower(), pk_k)
        if hasattr(self.cls, k):
            k = "%s_" % k
        if hasattr(self.cls, k):
            k_ = k
            i = 2
            while hasattr(self.cls, k):
                k = "%s_%i" % (k_, i)
                i += 1
        setattr(self.cls, k, c)
        c.validate()
        return self.get_relationship_with_model(model)

    def attrs_on_before_write(self, instance):
        attrs = self.get_data_attributes()
        for attr in attrs:
            attr.on_before_write(instance)

    def attrs_on_after_write(self, instance):
        attrs = self.get_data_attributes()
        for attr in attrs:
            attr.on_after_write(instance)

    def attrs_on_before_delete(self, instance, hard=None):
        attrs = self.get_data_attributes()
        for attr in attrs:
            attr.on_before_delete(instance, hard)

    def attrs_on_after_delete(self, instance, hard=None):
        attrs = self.get_data_attributes()
        for attr in attrs:
            attr.on_after_delete(instance, hard)





import copy
from pc23 import metaclass
from .meta import OrmModelTemplateMeta
from ..mapper import mapper
from ..attributes.base import AttributeAbstract
from ..exceptions import OrmParameterError


def with_model_template(*args):
    """ Decorator applying ORM templates to the end Model """
    def wrapper(cls):
        declaration_order = 0
        extra_attrs = list()
        keys_list = list()
        meta_data = dict()
        for template in args:
            if not issubclass(template, OrmModelTemplate):
                raise OrmParameterError(
                    "'with_model_template' requires that given Model's be a type of OrmModelTemplate!"
                )
            t_keys = dir(template)
            for k in t_keys:
                if k.startswith('__') and k.endswith('__'):
                    k_meta_attr = k[2:-2]
                    if hasattr(cls.__meta__, k_meta_attr) and not hasattr(cls, k) and k_meta_attr not in meta_data:
                        meta_data[k_meta_attr] = getattr(template, k)
                    continue
                if k in keys_list:
                    continue
                keys_list.append(k)
                c = getattr(template, k)
                extra_attrs.append((k, c, getattr(c, 'declaration_order', -1), template))
            mapper.register_templated_model(template, cls)
        extra_attrs = sorted(extra_attrs, key=lambda x: x[2])
        for attr in extra_attrs:
            k, co, decl_order, template = attr
            if hasattr(cls, k) and not getattr(getattr(cls, k), 'auto_created', False):
                continue
            c = copy.deepcopy(co)
            if isinstance(c, AttributeAbstract):
                c.declaration_order = declaration_order
                declaration_order += 1
                c.template_model = template
                c.template_k = k
            cls.__settemplatedattr__(k, c)
        for k in meta_data:
            setattr(cls.__meta__, k, meta_data[k])
            setattr(cls, '__%s__' % k, meta_data[k])
        return cls
    return wrapper


def use_model_template(base_model, *args):
    return with_model_template(*args)(base_model)


@metaclass(OrmModelTemplateMeta)
class OrmModelTemplate(object):
    pass


